"""
QUndoCommand class for generic undoable user actions performed to a BaseComponent

See `../life.py` for an example of a component that uses a custom QUndoCommand
"""

from PyQt6.QtGui import QUndoCommand
from copy import copy
import logging

log = logging.getLogger("AVP.ComponentHandler")


class ComponentUpdate(QUndoCommand):
    """Command object for making a component action undoable"""

    def __init__(self, parent, oldWidgetVals, modifiedVals):
        super().__init__("change %s component #%s" % (parent.name, parent.compPos))
        self.undone = False
        self.res = (int(parent.width), int(parent.height))
        self.parent = parent
        self.oldWidgetVals = {
            attr: (
                copy(val)
                if attr not in self.parent._relativeWidgets
                else self.parent.floatValForAttr(attr, val, axis=self.res)
            )
            for attr, val in oldWidgetVals.items()
            if attr in modifiedVals
        }
        self.modifiedVals = {
            attr: (
                val
                if attr not in self.parent._relativeWidgets
                else self.parent.floatValForAttr(attr, val, axis=self.res)
            )
            for attr, val in modifiedVals.items()
        }

        # Because relative widgets change themselves every update based on
        # their previous value, we must store ALL their values in case of undo
        self.relativeWidgetValsAfterUndo = {
            attr: copy(getattr(self.parent, attr))
            for attr in self.parent._relativeWidgets
        }

        # Determine if this update is mergeable
        self.id_ = -1
        if self.parent.mergeUndo:
            if len(self.modifiedVals) == 1:
                attr, val = self.modifiedVals.popitem()
                self.id_ = sum([ord(letter) for letter in attr[-14:]])
                self.modifiedVals[attr] = val
                return
            log.warning(
                "%s component settings changed at once. (%s)",
                len(self.modifiedVals),
                repr(self.modifiedVals),
            )

    def id(self):
        """If 2 consecutive updates have same id, Qt will call mergeWith()"""
        return self.id_

    def mergeWith(self, other):
        self.modifiedVals.update(other.modifiedVals)
        return True

    def setWidgetValues(self, attrDict):
        """
        Mask the component's usual method to handle our
        relative widgets in case the resolution has changed.
        """
        newAttrDict = {
            attr: (
                val
                if attr not in self.parent._relativeWidgets
                else self.parent.pixelValForAttr(attr, val)
            )
            for attr, val in attrDict.items()
        }
        self.parent.setWidgetValues(newAttrDict)

    def redo(self):
        if self.undone:
            log.info("Redoing component update")
        self.parent.oldAttrs = self.relativeWidgetValsAfterUndo
        self.setWidgetValues(self.modifiedVals)
        self.parent.update(auto=True)
        self.parent.oldAttrs = None
        if not self.undone:
            self.relativeWidgetValsAfterRedo = {
                attr: copy(getattr(self.parent, attr))
                for attr in self.parent._relativeWidgets
            }
            self.parent._sendUpdateSignal()

    def undo(self):
        log.info("Undoing component update")
        self.undone = True
        self.parent.oldAttrs = self.relativeWidgetValsAfterRedo
        self.setWidgetValues(self.oldWidgetVals)
        self.parent.update(auto=True)
        self.parent.oldAttrs = None



import sys
import pytest
from pytestqt import qtbot
from . import command


def test_commandline_help(qtbot, command):
    sys.argv = ["", "--help"]
    with pytest.raises(SystemExit):
        command.parseArgs()


def test_commandline_help_if_bad_args(qtbot, command):
    sys.argv = ["", "--junk"]
    with pytest.raises(SystemExit):
        command.parseArgs()


def test_commandline_launches_gui_if_verbose(qtbot, command):
    sys.argv = ["", "--verbose"]
    mode = command.parseArgs()
    assert mode == "GUI"


def test_commandline_launches_gui_if_verbose_with_project(qtbot, command):
    sys.argv = ["", "test", "--verbose"]
    mode = command.parseArgs()
    assert mode == "GUI"


def test_commandline_tries_to_export(qtbot, command):
    didCallFunction = False

    def captureFunction(*args):
        nonlocal didCallFunction
        didCallFunction = True

    sys.argv = ["", "-c", "0", "classic", "-i", "_", "-o", "_"]
    command.createAudioVisualization = captureFunction
    command.parseArgs()
    assert didCallFunction


def test_commandline_parses_classic_by_alias(qtbot, command):
    assert command.parseCompName("original") == "Classic Visualizer"


def test_commandline_parses_conway_by_short_name(qtbot, command):
    assert command.parseCompName("conway") == "Conway's Game of Life"


def test_commandline_parses_image_by_name(qtbot, command):
    assert command.parseCompName("image") == "Image"



import os
import tempfile
import numpy

from avp.core import Core
from avp.command import Command
from avp.gui.mainwindow import MainWindow
from avp.toolkit.ffmpeg import readAudioFile
from pytest import fixture


PYTEST_XDIST_WORKER_COUNT = os.environ.get("PYTEST_XDIST_WORKER_COUNT", 0)


@fixture
def settings():
    """Doesn't instantiate core: just calls a static method to store `settings.ini`"""
    initCore()
    yield None


@fixture
def audioData():
    """Fixture that gives a tuple of (completeAudioArray, duration)"""
    initCore()
    soundFile = getTestDataPath("inputfiles/test.ogg")
    yield readAudioFile(soundFile, MockVideoWorker())


@fixture
def command(qtbot):
    initCore()
    command = Command()
    command.quit = lambda _: None
    yield command


@fixture
def window(qtbot):
    initCore()
    # patch out any modal dialog that might happen
    MainWindow.showMessage = lambda self, msg, **kwargs: print(msg)
    window = MainWindow(None, None)
    window.clear()
    qtbot.addWidget(window)
    window.settings.setValue("outputWidth", 1920)
    window.settings.setValue("outputHeight", 1080)
    yield window


def getTestDataPath(filename=""):
    """Get path to a file in the ./data directory"""
    tests_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(tests_dir, "data", filename)


def initCore():
    """
    Initializes the Core by creating `settings.ini`
    Returns the temp directory path where settings.ini was created
    or None if multiple pytest workers are not enabled.
    """
    try:
        numWorkers = int(PYTEST_XDIST_WORKER_COUNT)
    except ValueError:
        numWorkers = 0
    if numWorkers > 0:
        # use temporary directories for multiple workers
        # so they don't interfere with each other
        configDir = tempfile.mkdtemp(prefix="avp-config-")
    else:
        # use test data path so we can easily see it after
        # a failed test, and help us understand the config
        configDir = getTestDataPath("config")
    unwanted = ["autosave.avp", "settings.ini"]
    for file in unwanted:
        filename = os.path.join(configDir, "autosave.avp")
        if os.path.exists(filename):
            os.remove(filename)
    Core.storeSettings(configDir)
    return configDir if numWorkers > 0 else None


def preFrameRender(audioData, comp):
    """Prepares a component for calls to frameRender()"""
    comp.preFrameRender(
        audioFile=getTestDataPath("inputfiles/test.ogg"),
        completeAudioArray=audioData[0],
        sampleSize=1470,
        progressBarSetText=MockSignal(),
        progressBarUpdate=MockSignal(),
    )


class MockSignal:
    """Pretends to be a pyqtSignal"""

    def emit(self, *args):
        pass


class MockVideoWorker:
    """Pretends to be a video thread worker"""

    progressBarSetText = MockSignal()
    progressBarUpdate = MockSignal()


def imageDataSum(image):
    """Get sum of raw data of a Pillow Image object"""
    return numpy.asarray(image, dtype="int32").sum(dtype="int32")



from PyQt6.QtWidgets import QApplication
import sys
import logging
import re
import string


from .core import Core

# Core class must store settings as class variables
# before we can use many things in AVP
Core.storeSettings()
log = logging.getLogger("AVP.Main")


def main() -> int:
    """Returns an exit code (0 for success)"""
    proj = None
    mode = "GUI"

    # Determine whether we're in GUI or commandline mode
    if len(sys.argv) > 2:
        mode = "commandline"
    elif len(sys.argv) == 2:
        if sys.argv[1].startswith("-"):
            mode = "commandline"
        else:
            # remove unsafe punctuation characters such as \/?*&^%$#
            if sys.argv[1].endswith(".avp"):
                # remove file extension
                sys.argv[1] = sys.argv[1][:-4]
            sys.argv[1] = re.sub(f"[{re.escape(string.punctuation)}]", "", sys.argv[1])
            # opening a project file with gui
            proj = sys.argv[1]

    # Create Qt Application
    app = QApplication(sys.argv)
    app.setApplicationName("audio-visualizer")

    screen = app.primaryScreen()
    if screen is None:
        dpi = None
    else:
        dpi = screen.physicalDotsPerInchX()

    # Launch program
    if mode == "commandline":
        from .command import Command

        main = Command()
        mode = main.parseArgs()
        log.debug("Finished creating command object")

    log.info(f"QApplication Platform: {QApplication.platformName()}")
    log.info(f"Detected screen DPI: {dpi}")
    # Both branches here may occur in one execution:
    # Commandline parsing could change mode back to GUI
    if mode == "GUI":
        from avp.gui.mainwindow import MainWindow

        mainWindow = MainWindow(proj, dpi)
        log.debug("Finished creating MainWindow")
        mainWindow.raise_()

    return app.exec()


if __name__ == "__main__":
    sys.exit(main())



from PyQt6 import QtCore, QtWidgets
from PyQt6.QtGui import QUndoCommand
from PIL import Image, ImageDraw
import os
import math
import logging


from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame, scale, addShadow
from ..toolkit.visualizer import createSpectrumArray


log = logging.getLogger("AVP.Component.Life")


class Component(BaseComponent):
    name = "Conway's Game of Life"
    version = "2.0.1"

    def widget(self, *args):
        super().widget(*args)
        self.scale = 32
        self.updateGridSize()
        # The initial grid: a "Queen Bee Shuttle"
        # https://conwaylife.com/wiki/Queen_bee_shuttle
        self.startingGrid = set(
            [
                (3, 11),
                (3, 12),
                (4, 11),
                (4, 12),
                (8, 11),
                (9, 10),
                (9, 12),
                (10, 9),
                (10, 13),
                (11, 10),
                (11, 11),
                (11, 12),
                (12, 8),
                (12, 9),
                (12, 13),
                (12, 14),
                (23, 10),
                (23, 11),
                (24, 10),
                (24, 11),
            ]
        )

        # Amount of 'bleed' (off-canvas coordinates) on each side of the grid
        self.bleedSize = 40

        self.page.pushButton_pickImage.clicked.connect(self.pickImage)
        self.trackWidgets(
            {
                "tickRate": self.page.spinBox_tickRate,
                "scale": self.page.spinBox_scale,
                "color": self.page.lineEdit_color,
                "shapeType": self.page.comboBox_shapeType,
                "shadow": self.page.checkBox_shadow,
                "customImg": self.page.checkBox_customImg,
                "showGrid": self.page.checkBox_showGrid,
                "image": self.page.lineEdit_image,
                "kaleidoscope": self.page.checkBox_kaleidoscope,
                "sensitivity": self.page.spinBox_sensitivity,
            },
            colorWidgets={
                "color": self.page.pushButton_color,
            },
        )
        self.shiftButtons = (
            self.page.toolButton_up,
            self.page.toolButton_down,
            self.page.toolButton_left,
            self.page.toolButton_right,
        )

        def shiftFunc(i):
            def shift():
                self.shiftGrid(i)

            return shift

        shiftFuncs = [shiftFunc(i) for i in range(len(self.shiftButtons))]
        for i, widget in enumerate(self.shiftButtons):
            widget.clicked.connect(shiftFuncs[i])
        self.page.spinBox_scale.setValue(self.scale)
        self.page.spinBox_scale.valueChanged.connect(self.updateGridSize)

    def pickImage(self):
        imgDir = self.settings.value("componentDir", os.path.expanduser("~"))
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self.page,
            "Choose Image",
            imgDir,
            "Image Files (%s)" % " ".join(self.core.imageFormats),
        )
        if filename:
            self.settings.setValue("componentDir", os.path.dirname(filename))
            self.mergeUndo = False
            self.page.lineEdit_image.setText(filename)
            self.mergeUndo = True

    def shiftGrid(self, d):
        action = ShiftGrid(self, d)
        self.parent.undoStack.push(action)

    def update(self):
        self.updateGridSize()

        # Hide/show widgets depending on state of "custom image" checkbox
        if self.page.checkBox_customImg.isChecked():
            self.page.label_color.setVisible(False)
            self.page.lineEdit_color.setVisible(False)
            self.page.pushButton_color.setVisible(False)
            self.page.label_shape.setVisible(False)
            self.page.comboBox_shapeType.setVisible(False)
            self.page.label_image.setVisible(True)
            self.page.lineEdit_image.setVisible(True)
            self.page.pushButton_pickImage.setVisible(True)
        else:
            self.page.label_color.setVisible(True)
            self.page.lineEdit_color.setVisible(True)
            self.page.pushButton_color.setVisible(True)
            self.page.label_shape.setVisible(True)
            self.page.comboBox_shapeType.setVisible(True)
            self.page.label_image.setVisible(False)
            self.page.lineEdit_image.setVisible(False)
            self.page.pushButton_pickImage.setVisible(False)

        # Disable audio sensitivity spinbox if not relevant
        if (
            self.page.comboBox_shapeType.currentIndex() < 4
            or self.page.checkBox_customImg.isChecked()
        ):
            self.page.spinBox_sensitivity.setEnabled(True)
        else:
            self.page.spinBox_sensitivity.setEnabled(False)

        # Disable arrow buttons to shift the grid if the grid is empty
        enabled = len(self.startingGrid) > 0
        for widget in self.shiftButtons:
            widget.setEnabled(enabled)

    def previewClickEvent(self, pos, size, button):
        pos = (
            math.ceil((pos[0] / size[0]) * self.gridWidth) - 1,
            math.ceil((pos[1] / size[1]) * self.gridHeight) - 1,
        )
        action = ClickGrid(self, pos, button)
        self.parent.undoStack.push(action)

    def updateGridSize(self):
        w, h = self.core.resolutions[-1].split("x")
        self.gridWidth = int(int(w) / self.scale)
        self.gridHeight = int(int(h) / self.scale)
        self.pxWidth = math.ceil(self.width / self.gridWidth)
        self.pxHeight = math.ceil(self.height / self.gridHeight)

    def previewRender(self):
        image = self.drawGrid(self.startingGrid, self.color)
        image = self.addKaleidoscopeEffect(image)
        if self.shadow:
            image = addShadow(image, 5.00, -2, 2)
        image = self.addGridLines(image)
        return image

    def preFrameRender(self, *args, **kwargs):
        super().preFrameRender(*args, **kwargs)
        self.tickGrids = {0: self.startingGrid}
        if self.sensitivity == 0:
            return

        self.spectrumArray = createSpectrumArray(
            self,
            self.completeAudioArray,
            self.sampleSize,
            0.08,
            0.8,
            20,
            self.progressBarUpdate,
            self.progressBarSetText,
        )

    def properties(self):
        if self.customImg and (not self.image or not os.path.exists(self.image)):
            return ["error"]
        return ["pcm"] if self.sensitivity > 0 else []

    def error(self):
        return "No image selected to represent life."

    def frameRender(self, frameNo):
        tick = math.floor(frameNo / self.tickRate)

        # Compute grid evolution on this frame if it hasn't been computed yet
        if tick not in self.tickGrids:
            self.tickGrids[tick] = self.gridForTick(tick)
        grid = self.tickGrids[tick]

        # Delete old evolution data which we shouldn't need anymore
        if tick - 60 in self.tickGrids:
            del self.tickGrids[tick - 60]

        # Fade difference between previous and current grid
        previousGrid = self.tickGrids.get(tick - 1, set())
        newColor = self.color
        if not self.customImg:
            r, g, b = self.color
            decay = 255 / self.tickRate
            decayAmount = int(decay * (frameNo % self.tickRate))
            decayColor = (
                r,
                g,
                b,
                255 - decayAmount,
            )
            newColor = (r, g, b, min(255, decayAmount * 2))
            previousGridImage = self.drawGrid(
                previousGrid,
                decayColor,
                (
                    None
                    if (not self.customImg and self.shapeType > 3)
                    or self.sensitivity < 1
                    else self.spectrumArray[frameNo * self.sampleSize]
                ),
            )
        image = self.drawGrid(
            grid,
            newColor,
            (
                None
                if (not self.customImg and self.shapeType > 3) or self.sensitivity < 1
                else self.spectrumArray[frameNo * self.sampleSize]
            ),
            grid.intersection(previousGrid),
        )
        if not self.customImg:
            image = Image.alpha_composite(previousGridImage, image)
        image = self.addKaleidoscopeEffect(image)
        if self.shadow:
            image = addShadow(image, 5.00, -2, 2)
        image = self.addGridLines(image)
        return image

    def addGridLines(self, frame):
        if not self.showGrid:
            return frame

        drawer = ImageDraw.Draw(frame)
        w, h = scale(0.05, self.width, self.height, int)
        for x in range(self.pxWidth, self.width, self.pxWidth):
            drawer.rectangle(
                ((x, 0), (x + w, self.height)),
                fill=self.color,
            )
        for y in range(self.pxHeight, self.height, self.pxHeight):
            drawer.rectangle(
                ((0, y), (self.width, y + h)),
                fill=self.color,
            )
        return frame

    def addKaleidoscopeEffect(self, frame):
        if not self.kaleidoscope:
            return frame

        flippedImage = frame.transpose(Image.Transpose.FLIP_TOP_BOTTOM)
        frame.paste(flippedImage, (0, 0), mask=flippedImage)

        flippedImage = frame.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
        frame.paste(flippedImage, (0, 0), mask=flippedImage)

        return frame

    def drawGrid(self, grid, color, spectrumData=None, didntChange=None):
        frame = BlankFrame(self.width, self.height)
        if didntChange is None:
            # this set would contain cell coords that did not change
            # between the previous grid tick and this one
            didntChange = set()

        def drawCustomImg():
            try:
                img = Image.open(self.image)
            except Exception:
                return
            img = img.resize(
                (
                    (self.pxWidth + audioMorphWidth),
                    (self.pxHeight + audioMorphHeight),
                ),
                Image.Resampling.LANCZOS,
            )
            frame.paste(
                img,
                box=(
                    (drawPtX - (audioMorphWidth * 2)),
                    (drawPtY - (audioMorphHeight * 2)),
                ),
            )

        def drawShape(x, y):
            drawer = ImageDraw.Draw(frame)
            rect = (
                (drawPtX - audioMorphWidth, drawPtY - audioMorphHeight),
                (
                    drawPtX + self.pxWidth + audioMorphWidth,
                    drawPtY + self.pxHeight + audioMorphHeight,
                ),
            )
            shape = self.page.comboBox_shapeType.currentText().lower()
            thisCellColor = color if (x, y) not in didntChange else (*color[:3], 255)

            # Rectangle
            if shape == "rectangle":
                drawer.rectangle(rect, fill=thisCellColor)

            # Elliptical
            elif shape == "elliptical":
                drawer.ellipse(rect, fill=thisCellColor)

            tenthX, tenthY = scale(10, self.pxWidth, self.pxHeight, int)
            smallerShape = (
                (
                    drawPtX + tenthX + int(tenthX / 4) - int(audioMorphWidth / 2),
                    drawPtY + tenthY + int(tenthY / 2) - int(audioMorphHeight / 2),
                ),
                (
                    drawPtX
                    + self.pxWidth
                    - tenthX
                    - int(tenthX / 4)
                    + int(audioMorphWidth / 2),
                    drawPtY
                    + self.pxHeight
                    - (tenthY + int(tenthY / 2))
                    + int(audioMorphHeight / 2),
                ),
            )
            outlineShape = (
                (
                    drawPtX + int(tenthX / 4) - audioMorphWidth,
                    drawPtY + int(tenthY / 2) - audioMorphHeight,
                ),
                (
                    drawPtX + self.pxWidth - int(tenthX / 4) + audioMorphWidth,
                    drawPtY + self.pxHeight - int(tenthY / 2) + audioMorphHeight,
                ),
            )
            # Circle
            if shape == "circle":
                drawer.ellipse(outlineShape, fill=thisCellColor)
                drawer.ellipse(smallerShape, fill=(0, 0, 0, 0))

            # Lilypad
            elif shape == "lilypad":
                drawer.pieslice(smallerShape, 290, 250, fill=thisCellColor)

            # Pie
            elif shape == "pie":
                drawer.pieslice(outlineShape, 35, 320, fill=thisCellColor)

            hX, hY = scale(50, self.pxWidth, self.pxHeight, int)  # halfline
            tX, tY = scale(33, self.pxWidth, self.pxHeight, int)  # thirdline
            qX, qY = scale(20, self.pxWidth, self.pxHeight, int)  # quarterline

            # Path
            if shape == "path":
                drawer.ellipse(rect, fill=thisCellColor)
                rects = {
                    direction: False
                    for direction in (
                        "up",
                        "down",
                        "left",
                        "right",
                    )
                }
                for cell in self.nearbyCoords(x, y):
                    if cell not in grid:
                        continue
                    if cell[0] == x:
                        if cell[1] < y:
                            rects["up"] = True
                        if cell[1] > y:
                            rects["down"] = True
                    if cell[1] == y:
                        if cell[0] < x:
                            rects["left"] = True
                        if cell[0] > x:
                            rects["right"] = True

                for direction, rect in rects.items():
                    if rect:
                        if direction == "up":
                            sect = (
                                (drawPtX, drawPtY),
                                (drawPtX + self.pxWidth, drawPtY + hY),
                            )
                        elif direction == "down":
                            sect = (
                                (drawPtX, drawPtY + hY),
                                (
                                    drawPtX + self.pxWidth,
                                    drawPtY + self.pxHeight,
                                ),
                            )
                        elif direction == "left":
                            sect = (
                                (drawPtX, drawPtY),
                                (drawPtX + hX, drawPtY + self.pxHeight),
                            )
                        elif direction == "right":
                            sect = (
                                (drawPtX + hX, drawPtY),
                                (
                                    drawPtX + self.pxWidth,
                                    drawPtY + self.pxHeight,
                                ),
                            )
                        drawer.rectangle(sect, fill=thisCellColor)

            # Duck
            elif shape == "duck":
                duckHead = (
                    (drawPtX + qX, drawPtY + qY),
                    (drawPtX + int(qX * 3), drawPtY + int(tY * 2)),
                )
                duckBeak = (
                    (drawPtX + hX, drawPtY + qY),
                    (drawPtX + self.pxWidth + qX, drawPtY + int(qY * 3)),
                )
                duckWing = ((drawPtX, drawPtY + hY), rect[1])
                duckBody = (
                    (drawPtX + int(qX / 4), drawPtY + int(qY * 3)),
                    (drawPtX + int(tX * 2), drawPtY + self.pxHeight),
                )
                drawer.ellipse(duckBody, fill=thisCellColor)
                drawer.ellipse(duckHead, fill=thisCellColor)
                drawer.pieslice(duckWing, 130, 200, fill=thisCellColor)
                drawer.pieslice(duckBeak, 145, 200, fill=thisCellColor)

            # Peace
            elif shape == "peace":
                line = (
                    (
                        drawPtX + hX - int(tenthX / 2),
                        drawPtY + int(tenthY / 2),
                    ),
                    (
                        drawPtX + hX + int(tenthX / 2),
                        drawPtY + self.pxHeight - int(tenthY / 2),
                    ),
                )
                drawer.ellipse(outlineShape, fill=thisCellColor)
                drawer.ellipse(smallerShape, fill=(0, 0, 0, 0))
                drawer.rectangle(line, fill=thisCellColor)

                def slantLine(difference):
                    return (
                        (drawPtX + difference),
                        (drawPtY + self.pxHeight - qY),
                    ), (
                        (drawPtX + hX),
                        (drawPtY + hY),
                    )

                drawer.line(slantLine(qX), fill=thisCellColor, width=tenthX)
                drawer.line(
                    slantLine(self.pxWidth - qX), fill=thisCellColor, width=tenthX
                )

        for x, y in grid:
            drawPtX = x * self.pxWidth
            if drawPtX > self.width or drawPtX + self.pxWidth < 0:
                continue
            drawPtY = y * self.pxHeight
            if drawPtY > self.height or drawPtY + self.pxHeight < 0:
                continue

            audioMorphWidth = (
                0 if spectrumData is None else int(spectrumData[(drawPtX % 63) * 4] / 4)
            )
            audioMorphHeight = (
                0 if spectrumData is None else int(spectrumData[(drawPtY % 63) * 4] / 4)
            )
            if self.customImg:
                drawCustomImg()
            else:
                drawShape(x, y)

        return frame

    def gridForTick(self, tick):
        """
        Given a tick number over 0, returns a new grid (a set of tuples).
        This must compute the previous ticks' grids if not already computed
        """
        if tick - 1 not in self.tickGrids:
            self.tickGrids[tick - 1] = self.gridForTick(tick - 1)

        lastGrid = self.tickGrids[tick - 1]

        def neighbours(x, y):
            return {cell for cell in self.nearbyCoords(x, y) if cell in lastGrid}

        newGrid = set()
        # Copy cells from the previous grid if they have 2 or 3 neighbouring cells
        # and if they are within the grid or its bleed area (off-canvas area)
        for x, y in lastGrid:
            if (
                -self.bleedSize > x > self.gridWidth + self.bleedSize
                or -self.bleedSize > y > self.gridHeight + self.bleedSize
            ):
                continue
            surrounding = len(neighbours(x, y))
            if surrounding == 2 or surrounding == 3:
                newGrid.add((x, y))

        # Find positions around living cells which must be checked for reproduction
        potentialNewCells = {
            coordTup
            for origin in lastGrid
            for coordTup in list(self.nearbyCoords(*origin))
        }
        # Check for reproduction
        for x, y in potentialNewCells:
            if (x, y) in newGrid:
                # Ignore non-empty cell
                continue
            surrounding = len(neighbours(x, y))
            if surrounding == 3:
                newGrid.add((x, y))

        return newGrid

    def savePreset(self):
        pr = super().savePreset()
        pr["GRID"] = sorted(self.startingGrid)
        return pr

    def loadPreset(self, pr, *args):
        self.startingGrid = set(pr["GRID"])
        if self.startingGrid:
            for widget in self.shiftButtons:
                widget.setEnabled(True)
        super().loadPreset(pr, *args)

    def nearbyCoords(self, x, y):
        yield x + 1, y + 1
        yield x + 1, y - 1
        yield x - 1, y + 1
        yield x - 1, y - 1
        yield x, y + 1
        yield x, y - 1
        yield x + 1, y
        yield x - 1, y


class ClickGrid(QUndoCommand):
    def __init__(self, comp, pos, button):
        super().__init__("click %s component #%s" % (comp.name, comp.compPos))
        self.comp = comp
        self.pos = [pos]
        if button == QtCore.Qt.MouseButton.RightButton:
            self.button = 2
        else:
            self.button = 1

    def id(self):
        return self.button

    def mergeWith(self, other):
        self.pos.extend(other.pos)
        return True

    def add(self):
        for pos in self.pos[:]:
            self.comp.startingGrid.add(pos)
        self.comp.update(auto=True)

    def remove(self):
        for pos in self.pos[:]:
            self.comp.startingGrid.discard(pos)
        self.comp.update(auto=True)

    def redo(self):
        if self.button == 1:  # Left-click
            self.add()
        elif self.button == 2:  # Right-click
            self.remove()

    def undo(self):
        if self.button == 1:  # Left-click
            self.remove()
        elif self.button == 2:  # Right-click
            self.add()


class ShiftGrid(QUndoCommand):
    def __init__(self, comp, direction):
        super().__init__("change %s component #%s" % (comp.name, comp.compPos))
        self.comp = comp
        self.direction = direction
        self.distance = 1

    def id(self):
        return self.direction

    def mergeWith(self, other):
        self.distance += other.distance
        return True

    def newGrid(self, Xchange, Ychange):
        return {(x + Xchange, y + Ychange) for x, y in self.comp.startingGrid}

    def redo(self):
        if self.direction == 0:
            newGrid = self.newGrid(0, -self.distance)
        elif self.direction == 1:
            newGrid = self.newGrid(0, self.distance)
        elif self.direction == 2:
            newGrid = self.newGrid(-self.distance, 0)
        elif self.direction == 3:
            newGrid = self.newGrid(self.distance, 0)
        self.comp.startingGrid = newGrid
        self.comp._sendUpdateSignal()

    def undo(self):
        if self.direction == 0:
            newGrid = self.newGrid(0, self.distance)
        elif self.direction == 1:
            newGrid = self.newGrid(0, -self.distance)
        elif self.direction == 2:
            newGrid = self.newGrid(self.distance, 0)
        elif self.direction == 3:
            newGrid = self.newGrid(-self.distance, 0)
        self.comp.startingGrid = newGrid
        self.comp._sendUpdateSignal()



from avp.command import Command
from PyQt6.QtGui import QFont
from pytestqt import qtbot
from pytest import fixture, mark
from . import audioData, command, MockSignal, imageDataSum


@fixture
def coreWithTextComp(qtbot, command):
    """Fixture providing a Command object with Title Text component added"""
    command.core.insertComponent(0, command.core.moduleIndexFor("Title Text"), command)
    yield command.core


def setTextSettings(comp):
    comp.page.spinBox_fontSize.setValue(40)
    comp.page.checkBox_shadow.setChecked(True)
    comp.page.spinBox_shadBlur.setValue(0)
    comp.page.spinBox_shadX.setValue(2)
    comp.page.spinBox_shadY.setValue(-2)
    comp.page.fontComboBox_titleFont.setCurrentFont(QFont("Noto Sans"))
    comp.page.lineEdit_textColor.setText("255,255,255")


@mark.parametrize(
    "width, height",
    ((1920, 1080), (1280, 720)),
)
def test_comp_text_renderFrame(coreWithTextComp, width, height):
    """Call renderFrame of Title Text component added to Command object."""
    comp = coreWithTextComp.selectedComponents[0]
    comp.parent.settings.setValue("outputWidth", width)
    comp.parent.settings.setValue("outputHeight", height)
    setTextSettings(comp)
    comp.centerXY()
    image = comp.frameRender(0)
    assert comp.titleFont.family() == "Noto Sans"
    assert comp.xPosition == width / 2
    assert image.width == width
    assert comp.fontSize == 40
    assert comp.shadX == 2
    assert comp.shadY == -2
    assert comp.shadBlur == 0
    assert imageDataSum(image) == 727403 or 738586



"""
QUndoCommand classes for every undoable user action performed in the MainWindow
"""

from PyQt6.QtGui import QUndoCommand
import os
import logging
from copy import copy

from ..core import Core


log = logging.getLogger("AVP.Gui.Actions")


# =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
# COMPONENT ACTIONS
# =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


class AddComponent(QUndoCommand):
    def __init__(self, parent, compI, moduleI):
        super().__init__(
            "create new %s component" % parent.core.modules[moduleI].Component.name
        )
        self.parent = parent
        self.moduleI = moduleI
        self.compI = compI
        self.comp = None
        self.valid = True

    def redo(self):
        if self.comp is None:
            i = self.parent.core.insertComponent(self.compI, self.moduleI, self.parent)
            if i != self.compI:
                self.valid = False
                if i is not None:
                    log.error(
                        f"Expected new component index to be {self.compI} but received {i}"
                    )
        else:
            # inserting previously-created component
            self.parent.core.insertComponent(self.compI, self.comp, self.parent)

    def undo(self):
        if not self.valid:
            return
        self.comp = self.parent.core.selectedComponents[self.compI]
        self.parent._removeComponent(self.compI)


class RemoveComponent(QUndoCommand):
    def __init__(self, parent, selectedRows):
        super().__init__("remove component")
        self.parent = parent
        componentList = self.parent.listWidget_componentList
        self.selectedRows = [componentList.row(selected) for selected in selectedRows]
        self.components = [parent.core.selectedComponents[i] for i in self.selectedRows]

    def redo(self):
        self.parent._removeComponent(self.selectedRows[0])

    def undo(self):
        componentList = self.parent.listWidget_componentList
        for index, comp in zip(self.selectedRows, self.components):
            self.parent.core.insertComponent(index, comp, self.parent)
        self.parent.drawPreview()


class MoveComponent(QUndoCommand):
    def __init__(self, parent, row, newRow, tag):
        super().__init__("move component %s" % tag)
        self.parent = parent
        self.row = row
        self.newRow = newRow
        self.id_ = ord(tag[0])

    def id(self):
        """If 2 consecutive updates have same id, Qt will call mergeWith()"""
        return self.id_

    def mergeWith(self, other):
        self.newRow = other.newRow
        return True

    def do(self, rowa, rowb):
        componentList = self.parent.listWidget_componentList

        page = self.parent.pages.pop(rowa)
        self.parent.pages.insert(rowb, page)

        item = componentList.takeItem(rowa)
        componentList.insertItem(rowb, item)

        stackedWidget = self.parent.stackedWidget
        widget = stackedWidget.removeWidget(page)
        stackedWidget.insertWidget(rowb, page)
        componentList.setCurrentRow(rowb)
        stackedWidget.setCurrentIndex(rowb)
        self.parent.core.moveComponent(rowa, rowb)
        self.parent.drawPreview(True)

    def redo(self):
        self.do(self.row, self.newRow)

    def undo(self):
        self.do(self.newRow, self.row)


# =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
# PRESET ACTIONS
# =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


class ClearPreset(QUndoCommand):
    def __init__(self, parent, compI):
        super().__init__("clear preset")
        self.parent = parent
        self.compI = compI
        self.component = self.parent.core.selectedComponents[compI]
        self.store = self.component.savePreset()
        self.store["preset"] = self.component.currentPreset

    def redo(self):
        self.parent.core.clearPreset(self.compI)
        self.parent.updateComponentTitle(self.compI, False)

    def undo(self):
        self.parent.core.selectedComponents[self.compI].loadPreset(self.store)
        self.parent.updateComponentTitle(self.compI, self.store)


class OpenPreset(QUndoCommand):
    def __init__(self, parent, presetName, compI):
        super().__init__("open %s preset" % presetName)
        self.parent = parent
        self.presetName = presetName
        self.compI = compI

        comp = self.parent.core.selectedComponents[compI]
        self.store = comp.savePreset()
        self.store["preset"] = copy(comp.currentPreset)

    def redo(self):
        self.parent._openPreset(self.presetName, self.compI)

    def undo(self):
        self.parent.core.selectedComponents[self.compI].loadPreset(self.store)
        self.parent.parent.updateComponentTitle(self.compI, self.store)


class RenamePreset(QUndoCommand):
    def __init__(self, parent, path, oldName, newName):
        super().__init__("rename preset")
        self.parent = parent
        self.path = path
        self.oldName = oldName
        self.newName = newName

    def redo(self):
        self.parent.renamePreset(self.path, self.oldName, self.newName)

    def undo(self):
        self.parent.renamePreset(self.path, self.newName, self.oldName)


class DeletePreset(QUndoCommand):
    def __init__(self, parent, compName, vers, presetFile):
        self.parent = parent
        self.preset = (compName, vers, presetFile)
        self.path = os.path.join(Core.presetDir, compName, str(vers), presetFile)
        self.store = self.parent.core.getPreset(self.path)
        self.presetName = self.store["preset"]
        super().__init__("delete %s preset (%s)" % (self.presetName, compName))
        self.loadedPresets = [
            i
            for i, comp in enumerate(self.parent.core.selectedComponents)
            if self.presetName == str(comp.currentPreset)
        ]

    def redo(self):
        os.remove(self.path)
        for i in self.loadedPresets:
            self.parent.core.clearPreset(i)
            self.parent.parent.updateComponentTitle(i, False)
        self.parent.findPresets()
        self.parent.drawPresetList()

    def undo(self):
        self.parent.createNewPreset(*self.preset, self.store)
        selectedComponents = self.parent.core.selectedComponents
        for i in self.loadedPresets:
            selectedComponents[i].currentPreset = self.presetName
            self.parent.parent.updateComponentTitle(i)
        self.parent.findPresets()
        self.parent.drawPresetList()



"""
Preset manager object handles all interactions with presets, including
the context menu accessed from MainWindow.
"""

from PyQt6 import QtCore, QtWidgets, uic
import string
import os
import logging

from ..toolkit import badName
from ..core import Core, appName
from .actions import *


log = logging.getLogger("AVP.Gui.PresetManager")


class PresetManager(QtWidgets.QDialog):
    def __init__(self, parent):
        super().__init__()
        uic.loadUi(os.path.join(Core.wd, "gui", "presetmanager.ui"), self)
        self.parent = parent
        self.core = parent.core
        self.settings = parent.settings
        self.presetDir = parent.presetDir
        if not self.settings.value("presetDir"):
            self.settings.setValue("presetDir", os.path.join(parent.dataDir, "presets"))

        self.findPresets()

        # window
        self.lastFilter = "*"
        self.presetRows = []  # list of (comp, vers, name) tuples

        self.setWindowFlags(QtCore.Qt.WindowType.WindowStaysOnTopHint)

        # connect button signals
        self.pushButton_delete.clicked.connect(self.openDeletePresetDialog)
        self.pushButton_rename.clicked.connect(self.openRenamePresetDialog)
        self.pushButton_import.clicked.connect(self.openImportDialog)
        self.pushButton_export.clicked.connect(self.openExportDialog)
        self.pushButton_close.clicked.connect(self.close)

        # create filter box and preset list
        self.drawFilterList()
        self.comboBox_filter.currentIndexChanged.connect(
            lambda: self.drawPresetList(
                self.comboBox_filter.currentText(), self.lineEdit_search.text()
            )
        )

        # make auto-completion for search bar
        self.autocomplete = QtCore.QStringListModel()
        completer = QtWidgets.QCompleter()
        completer.setModel(self.autocomplete)
        self.lineEdit_search.setCompleter(completer)
        self.lineEdit_search.textChanged.connect(
            lambda: self.drawPresetList(
                self.comboBox_filter.currentText(), self.lineEdit_search.text()
            )
        )
        self.drawPresetList("*")

    def show_(self):
        """Open a new preset manager window from the mainwindow"""
        self.findPresets()
        self.drawFilterList()
        self.drawPresetList("*")
        self.show()

    def findPresets(self):
        log.debug("Searching %s for presets", self.presetDir)
        parseList = []
        for dirpath, dirnames, filenames in os.walk(self.presetDir):
            # anything without a subdirectory must be a preset folder
            if dirnames:
                continue
            for preset in filenames:
                compName = os.path.basename(os.path.dirname(dirpath))
                if compName not in self.core.compNames:
                    continue
                compVers = os.path.basename(dirpath)
                try:
                    parseList.append((compName, int(compVers), preset))
                except ValueError:
                    continue
        self.presets = {
            compName: [
                (vers, preset) for name, vers, preset in parseList if name == compName
            ]
            for compName, _, __ in parseList
        }

    def drawPresetList(self, compFilter=None, presetFilter=""):
        self.listWidget_presets.clear()
        if compFilter:
            self.lastFilter = str(compFilter)
        else:
            compFilter = str(self.lastFilter)
        self.presetRows = []
        presetNames = []
        for component, presets in self.presets.items():
            if compFilter != "*" and component != compFilter:
                continue
            for vers, preset in presets:
                if not presetFilter or presetFilter in preset:
                    self.listWidget_presets.addItem("%s: %s" % (component, preset))
                    self.presetRows.append((component, vers, preset))
                if preset not in presetNames:
                    presetNames.append(preset)
        self.autocomplete.setStringList(presetNames)

    def drawFilterList(self):
        self.comboBox_filter.clear()
        self.comboBox_filter.addItem("*")
        for component in self.presets:
            self.comboBox_filter.addItem(component)

    def clearPreset(self, compI=None):
        """Functions on mainwindow level from the context menu"""
        compI = self.parent.listWidget_componentList.currentRow()
        action = ClearPreset(self.parent, compI)
        self.parent.undoStack.push(action)

    def openSavePresetDialog(self):
        """Functions on mainwindow level from the context menu"""
        selectedComponents = self.core.selectedComponents
        componentList = self.parent.listWidget_componentList

        if componentList.currentRow() == -1:
            return
        while True:
            index = componentList.currentRow()
            currentPreset = selectedComponents[index].currentPreset
            newName, OK = QtWidgets.QInputDialog.getText(
                self.parent,
                appName,
                "New Preset Name:",
                QtWidgets.QLineEdit.EchoMode.Normal,
                currentPreset,
            )
            if OK:
                if badName(newName):
                    self.warnMessage(self.parent)
                    continue
                if newName:
                    if index != -1:
                        selectedComponents[index].currentPreset = newName
                        saveValueStore = selectedComponents[index].savePreset()
                        saveValueStore["preset"] = newName
                        componentName = str(selectedComponents[index]).strip()
                        vers = selectedComponents[index].version
                        self.createNewPreset(
                            componentName,
                            vers,
                            newName,
                            saveValueStore,
                            window=self.parent,
                        )
                        self.findPresets()
                        self.drawPresetList()
                        self.openPreset(newName, index)
            break

    def createNewPreset(self, compName, vers, filename, saveValueStore, **kwargs):
        path = os.path.join(self.presetDir, compName, str(vers), filename)
        if self.presetExists(path, **kwargs):
            return
        self.core.createPresetFile(compName, vers, filename, saveValueStore)

    def presetExists(self, path, **kwargs):
        if os.path.exists(path):
            window = kwargs.get("window", self)
            ch = self.parent.showMessage(
                msg="%s already exists! Overwrite it?" % os.path.basename(path),
                showCancel=True,
                icon="Warning",
                parent=window,
            )
            if not ch:
                # user clicked cancel
                return True

        return False

    def openPreset(self, presetName, compPos=None):
        componentList = self.parent.listWidget_componentList
        index = compPos if compPos is not None else componentList.currentRow()
        if index == -1:
            return
        action = OpenPreset(self, presetName, index)
        self.parent.undoStack.push(action)

    def _openPreset(self, presetName, index):
        selectedComponents = self.core.selectedComponents

        componentName = selectedComponents[index].name.strip()
        version = selectedComponents[index].version
        dirname = os.path.join(self.presetDir, componentName, str(version))
        filepath = os.path.join(dirname, presetName)
        self.core.openPreset(filepath, index, presetName)

        self.parent.updateComponentTitle(index)
        self.parent.drawPreview()

    def openDeletePresetDialog(self):
        row = self.getPresetRow()
        if row == -1:
            return
        comp, vers, name = self.presetRows[row]
        ch = self.parent.showMessage(
            msg="Really delete %s?" % name,
            showCancel=True,
            icon="Warning",
            parent=self,
        )
        if not ch:
            return
        self.deletePreset(comp, vers, name)

    def deletePreset(self, comp, vers, name):
        action = DeletePreset(self, comp, vers, name)
        self.parent.undoStack.push(action)

    def warnMessage(self, window=None):
        self.parent.showMessage(
            msg="Preset names must contain only letters, " "numbers, and spaces.",
            parent=window if window else self,
        )

    def getPresetRow(self):
        row = self.listWidget_presets.currentRow()
        if row > -1:
            return row

        # check if component selected in MainWindow has preset loaded
        componentList = self.parent.listWidget_componentList
        compIndex = componentList.currentRow()
        if compIndex == -1:
            return compIndex

        preset = self.core.selectedComponents[compIndex].currentPreset
        if preset is None:
            return -1
        else:
            rowTuple = (
                self.core.selectedComponents[compIndex].name,
                self.core.selectedComponents[compIndex].version,
                preset,
            )
            for i, tup in enumerate(self.presetRows):
                if rowTuple == tup:
                    index = i
                    break
            else:
                return -1
        return index

    def openRenamePresetDialog(self):
        presetList = self.listWidget_presets
        index = self.getPresetRow()
        if index == -1:
            return

        while True:
            newName, OK = QtWidgets.QInputDialog.getText(
                self,
                "Preset Manager",
                "Rename Preset:",
                QtWidgets.QLineEdit.EchoMode.Normal,
                self.presetRows[index][2],
            )
            if OK:
                if badName(newName):
                    self.warnMessage()
                    continue
                if newName:
                    comp, vers, oldName = self.presetRows[index]
                    path = os.path.join(self.presetDir, comp, str(vers))
                    newPath = os.path.join(path, newName)
                    if self.presetExists(newPath):
                        return
                    action = RenamePreset(self, path, oldName, newName)
                    self.parent.undoStack.push(action)
            break

    def renamePreset(self, path, oldName, newName):
        oldPath = os.path.join(path, oldName)
        newPath = os.path.join(path, newName)
        if os.path.exists(newPath):
            os.remove(newPath)
        os.rename(oldPath, newPath)
        self.findPresets()
        self.drawPresetList()
        path = os.path.dirname(newPath)
        for i, comp in enumerate(self.core.selectedComponents):
            if self.core.getPresetDir(comp) == path and comp.currentPreset == oldName:
                self.core.openPreset(newPath, i, newName)
                self.parent.updateComponentTitle(i, False)
                self.parent.drawPreview()

    def openImportDialog(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            "Import Preset File",
            self.settings.value("presetDir"),
            "Preset Files (*.avl)",
        )
        if filename:
            # get installed path & ask user to overwrite if needed
            path = ""
            while True:
                if path:
                    if self.presetExists(path):
                        break
                    else:
                        if os.path.exists(path):
                            os.remove(path)
                success, path = self.core.importPreset(filename)
                if success:
                    break

            self.findPresets()
            self.drawPresetList()
            self.settings.setValue("presetDir", os.path.dirname(filename))

    def openExportDialog(self):
        index = self.getPresetRow()
        if index == -1:
            return
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Export Preset",
            self.settings.value("presetDir"),
            "Preset Files (*.avl)",
        )
        if filename:
            comp, vers, name = self.presetRows[index]
            if not self.core.exportPreset(filename, comp, vers, name):
                self.parent.showMessage(
                    msg="Couldn't export %s." % filename, parent=self
                )
            self.settings.setValue("presetDir", os.path.dirname(filename))

    def clearPresetListSelection(self):
        self.listWidget_presets.setCurrentRow(-1)



from PyQt6 import QtCore, QtGui, QtWidgets
import logging

log = logging.getLogger("AVP.Gui.PreviewWindow")


class PreviewWindow(QtWidgets.QLabel):
    """
    Paints the preview QLabel in MainWindow and maintains the aspect ratio
    when the window is resized.
    """

    def __init__(self, parent, img):
        super().__init__()
        self.parent = parent
        # FIXME
        # self.setFrameStyle(QtWidgets.QFrame.StyledPanel)
        self.pixmap = QtGui.QPixmap(img)

    def paintEvent(self, event):
        size = self.size()
        painter = QtGui.QPainter(self)
        point = QtCore.QPoint(0, 0)
        scaledPix = self.pixmap.scaled(
            size,
            QtCore.Qt.AspectRatioMode.KeepAspectRatio,
            transformMode=QtCore.Qt.TransformationMode.SmoothTransformation,
        )

        # start painting the label from left upper corner
        point.setX(int((size.width() - scaledPix.width()) / 2))
        point.setY(int((size.height() - scaledPix.height()) / 2))
        painter.drawPixmap(point, scaledPix)

    def changePixmap(self, img):
        self.pixmap = QtGui.QPixmap(img)
        self.repaint()

    def mousePressEvent(self, event):
        if self.parent.encoding:
            return

        i = self.parent.listWidget_componentList.currentRow()
        if i >= 0:
            component = self.parent.core.selectedComponents[i]
            if not hasattr(component, "previewClickEvent"):
                return
            qpoint = event.position().toPoint()
            pos = (qpoint.x(), qpoint.y())
            size = (self.width(), self.height())
            butt = event.button()
            log.info("Click event for #%s: %s button %s" % (i, pos, butt))
            component.previewClickEvent(pos, size, butt)

    @QtCore.pyqtSlot(str)
    def threadError(self, msg):
        self.parent.showMessage(msg=msg, icon="Critical", parent=self)
        log.info("%", repr(self.parent))



from PyQt6 import QtWidgets
import os

from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame


class Component(BaseComponent):
    name = "Sound"
    version = "1.0.0"

    def widget(self, *args):
        super().widget(*args)
        self.page.pushButton_sound.clicked.connect(self.pickSound)
        self.trackWidgets(
            {
                "sound": self.page.lineEdit_sound,
                "chorus": self.page.checkBox_chorus,
                "delay": self.page.spinBox_delay,
                "volume": self.page.spinBox_volume,
            },
            commandArgs={
                "sound": None,
            },
        )

    def properties(self):
        props = ["static", "audio"]
        if not os.path.exists(self.sound):
            props.append("error")
        return props

    def error(self):
        if not self.sound:
            return "No audio file selected."
        if not os.path.exists(self.sound):
            return "The audio file selected no longer exists!"

    def audio(self):
        params = {}
        if self.delay != 0.0:
            params["adelay"] = "=%s" % str(int(self.delay * 1000.00))
        if self.chorus:
            params["chorus"] = "=0.5:0.9:50|60|40:0.4|0.32|0.3:0.25|0.4|0.3:2|2.3|1.3"
        if self.volume != 1.0:
            params["volume"] = "=%s:replaygain_noclip=0" % str(self.volume)

        return (self.sound, params)

    def pickSound(self):
        sndDir = self.settings.value("componentDir", os.path.expanduser("~"))
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
            self.page,
            "Choose Sound",
            sndDir,
            "Audio Files (%s)" % " ".join(self.core.audioFormats),
        )
        if filename:
            self.settings.setValue("componentDir", os.path.dirname(filename))
            self.mergeUndo = False
            self.page.lineEdit_sound.setText(filename)
            self.mergeUndo = True

    def commandHelp(self):
        print("Path to audio file:\n    path=/filepath/to/sound.ogg")

    def command(self, arg):
        if "=" in arg:
            key, arg = arg.split("=", 1)
            if key == "path":
                if "*%s" % os.path.splitext(arg)[1] not in self.core.audioFormats:
                    print("Not a supported audio format")
                    quit(1)
                self.page.lineEdit_sound.setText(arg)
                return

        super().command(arg)



"""
Common tools for drawing compatible frames in a Component's frameRender()
"""

from PyQt6 import QtGui
from PIL import Image, ImageEnhance, ImageChops, ImageFilter
from PIL.ImageQt import ImageQt
from PyQt6 import QtCore
import sys
import os
import math
import logging
from .. import core


log = logging.getLogger("AVP.Toolkit.Frame")


class FramePainter(QtGui.QPainter):
    """
    A QPainter for a blank frame, which can be converted into a
    Pillow image with finalize()
    """

    def __init__(self, width, height):
        image = BlankFrame(width, height)
        log.debug("Creating QImage from PIL image object")
        self.image = ImageQt(image)
        super().__init__(self.image)

    def setPen(self, penStyle):
        if type(penStyle) is tuple:
            super().setPen(QtGui.QColor(*penStyle))
        else:
            super().setPen(penStyle)

    def finalize(self):
        log.verbose("Finalizing FramePainter")
        buffer = QtCore.QBuffer()
        buffer.open(QtCore.QBuffer.OpenModeFlag.ReadWrite)
        self.image.save(buffer, "PNG")
        import io

        frame = Image.open(io.BytesIO(buffer.data()))
        buffer.close()
        self.end()
        return frame


def addShadow(frame, blurRadius, blurOffsetX, blurOffsetY):
    shadImg = ImageEnhance.Contrast(frame).enhance(0.0)
    shadImg = shadImg.filter(ImageFilter.GaussianBlur(blurRadius))
    frame = shadImg.paste(frame, box=(-blurOffsetX, -blurOffsetY), mask=frame)
    frame = shadImg
    return frame


def scale(scalePercent, width, height, returntype=None):
    width = (float(width) / 100.0) * float(scalePercent)
    height = (float(height) / 100.0) * float(scalePercent)
    if returntype == str:
        return (str(math.ceil(width)), str(math.ceil(height)))
    elif returntype == int:
        return (math.ceil(width), math.ceil(height))
    else:
        return (width, height)


def defaultSize(framefunc):
    """Makes width/height arguments optional"""

    def decorator(*args):
        if len(args) < 2:
            newArgs = list(args)
            if len(args) == 0 or len(args) == 1:
                height = int(core.Core.settings.value("outputHeight"))
                newArgs.append(height)
            if len(args) == 0:
                width = int(core.Core.settings.value("outputWidth"))
                newArgs.insert(0, width)
            args = tuple(newArgs)
        return framefunc(*args)

    return decorator


def FloodFrame(width, height, RgbaTuple):
    return Image.new("RGBA", (width, height), RgbaTuple)


@defaultSize
def BlankFrame(width, height):
    """The base frame used by each component to start drawing."""
    return FloodFrame(width, height, (0, 0, 0, 0))


@defaultSize
def Checkerboard(width, height):
    """
    A checkerboard to represent transparency to the user.
    """
    # TODO: Would be cool to generate this image with numpy instead.
    log.debug("Creating new %s*%s checkerboard" % (width, height))
    image = FloodFrame(1920, 1080, (0, 0, 0, 0))
    image.paste(Image.open(os.path.join(core.Core.wd, "gui", "background.png")), (0, 0))
    image = image.resize((width, height))
    return image



"""
Base classes for components to import. Read comments for some documentation
on making a valid component.
"""

from PyQt6 import uic, QtCore, QtWidgets
from PyQt6.QtGui import QColor
import os
import math
import logging
from copy import copy

from .metaclass import ComponentMetaclass
from .actions import ComponentUpdate
from .exceptions import ComponentError
from ..toolkit.frame import BlankFrame

from ..toolkit import (
    getWidgetValue,
    setWidgetValue,
    rgbFromString,
    randomColor,
    blockSignals,
)

log = logging.getLogger("AVP.BaseComponent")


class Component(QtCore.QObject, metaclass=ComponentMetaclass):
    """
    The base class for components to inherit.
    """

    name = "Component"
    # ui = 'name_Of_Non_Default_Ui_File'

    version = "1.0.0"
    # The major version (before the first dot) is used to determine
    # preset compatibility; the rest is ignored so it can be non-numeric.

    modified = QtCore.pyqtSignal(int, dict)
    _error = QtCore.pyqtSignal(str, str)

    def __init__(self, moduleIndex, compPos, core):
        super().__init__()
        self.moduleIndex = moduleIndex
        self.compPos = compPos
        self.core = core

        # STATUS VARIABLES
        self.currentPreset = None
        self._allWidgets = {}
        self._trackedWidgets = {}
        self._presetNames = {}
        self._commandArgs = {}
        self._colorWidgets = {}
        self._colorFuncs = {}
        self._relativeWidgets = {}
        # Pixel values stored as floats
        self._relativeValues = {}
        # Maximum values of spinBoxes at 1080p (Core.resolutions[0])
        self._relativeMaximums = {}

        # LOCKING VARIABLES
        self.openingPreset = False
        self.mergeUndo = True
        self._lockedProperties = None
        self._lockedError = None
        self._lockedSize = None
        # If set to a dict, values are used as basis to update relative widgets
        self.oldAttrs = None
        # Stop lengthy processes in response to this variable
        self.canceled = False

    def __str__(self):
        return self.__class__.name

    def __repr__(self):
        import pprint

        try:
            preset = self.savePreset()
        except Exception as e:
            preset = "%s occurred while saving preset" % str(e)

        return "Component(module %s, pos %s) (%s)\n" "Name: %s v%s\nPreset: %s" % (
            self.moduleIndex,
            self.compPos,
            object.__repr__(self),
            self.__class__.name,
            str(self.__class__.version),
            pprint.pformat(preset),
        )

    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
    # Render Methods
    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

    def previewRender(self):
        image = BlankFrame(self.width, self.height)
        return image

    def preFrameRender(self, **kwargs):
        """
        Must call super() when subclassing
        Triggered only before a video is exported (video_thread.py)
            self.audioFile = filepath to the main input audio file
            self.completeAudioArray = a list of audio samples
            self.sampleSize = number of audio samples per video frame
            self.progressBarUpdate = signal to set progress bar number
            self.progressBarSetText = signal to set progress bar text
        Use the latter two signals to update the MainWindow if needed
        for a long initialization procedure (i.e., for a visualizer)
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

    def frameRender(self, frameNo):
        audioArrayIndex = frameNo * self.sampleSize
        image = BlankFrame(self.width, self.height)
        return image

    def postFrameRender(self):
        pass

    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
    # Properties
    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

    def properties(self):
        """
        Return a list of properties with certain meanings:
        `static`: non-animated
        `audio`: has extra sound to add
        `error`: bad configuration
        `pcm`: request raw audio data
        `composite`: request frame to draw on
        """
        return []

    def error(self):
        """
        Return a string containing an error message, or None for a default.
        Or tuple of two strings for a message with details.
        Alternatively use lockError(msgString) within properties()
        to skip this method entirely.
        """
        return

    def audio(self):
        """
        Return audio to mix into master as a tuple with two elements:
        The first element can be:
            - A string (path to audio file),
            - Or an object that returns audio data through a pipe
        The second element must be a dictionary of ffmpeg filters/options
        to apply to the input stream. See the filter docs for ideas:
        https://ffmpeg.org/ffmpeg-filters.html
        """

    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
    # Idle Methods
    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

    def widget(self, parent):
        """
        Call super().widget(*args) to create the component widget
        which also auto-connects any common widgets (e.g., checkBoxes)
        to self.update(). Then in a subclass connect special actions
        (e.g., pushButtons to select a file) and initialize
        """
        self.parent = parent
        self.settings = parent.settings
        log.verbose(
            "Creating UI for %s #%s's widget",
            self.__class__.name,
            self.compPos,
        )
        self.page = self.loadUi(self.__class__.ui)

        # Find all normal widgets which will be connected after subclass method
        self._allWidgets = {
            "lineEdit": self.page.findChildren(QtWidgets.QLineEdit),
            "checkBox": self.page.findChildren(QtWidgets.QCheckBox),
            "spinBox": self.page.findChildren(QtWidgets.QSpinBox),
            "comboBox": self.page.findChildren(QtWidgets.QComboBox),
        }
        self._allWidgets["spinBox"].extend(
            self.page.findChildren(QtWidgets.QDoubleSpinBox)
        )

    def update(self):
        """
        Starting point for a component update. A subclass should override
        this method, and the base class will then magically insert a call
        to either _autoUpdate() or _userUpdate() at the end.
        """

    def loadPreset(self, presetDict, presetName=None):
        """
        Subclasses should take (presetDict, *args) as args.
        Must use super().loadPreset(presetDict, *args) first,
        then update self.page widgets using the preset dict.
        """
        self.currentPreset = (
            presetName if presetName is not None else presetDict["preset"]
        )
        for attr, widget in self._trackedWidgets.items():
            key = attr if attr not in self._presetNames else self._presetNames[attr]
            try:
                val = presetDict[key]
            except KeyError as e:
                log.info(
                    "%s missing value %s. Outdated preset?",
                    self.currentPreset,
                    str(e),
                )
                val = getattr(self, key)

            if attr in self._colorWidgets:
                widget.setText("%s,%s,%s" % val)
                btnStyle = (
                    "QPushButton { background-color : %s; outline: none; }"
                    % QColor(*val).name()
                )
                self._colorWidgets[attr].setStyleSheet(btnStyle)
            elif attr in self._relativeWidgets:
                self._relativeValues[attr] = val
                pixelVal = self.pixelValForAttr(attr, val)
                setWidgetValue(widget, pixelVal)
            else:
                setWidgetValue(widget, val)

    def savePreset(self):
        saveValueStore = {}
        for attr, widget in self._trackedWidgets.items():
            presetAttrName = (
                attr if attr not in self._presetNames else self._presetNames[attr]
            )
            if attr in self._relativeWidgets:
                try:
                    val = self._relativeValues[attr]
                except AttributeError:
                    val = self.floatValForAttr(attr)
            else:
                val = getattr(self, attr)

            saveValueStore[presetAttrName] = val
        return saveValueStore

    def commandHelp(self):
        """Help text as string for this component's commandline arguments"""

    def command(self, arg=""):
        """
        Configure a component using an arg from the commandline. This is
        never called if global args like 'preset=' are found in the arg.
        So simply check for any non-global args in your component and
        call super().command() at the end to get a Help message.
        """
        print(
            self.__class__.name,
            "Usage:\n" "Open a preset for this component:\n" '    "preset=Preset Name"',
        )
        self.commandHelp()
        quit(0)

    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
    # "Private" Methods
    # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
    def _preUpdate(self):
        """Happens before subclass update()"""
        for attr in self._relativeWidgets:
            self.updateRelativeWidget(attr)

    def _userUpdate(self):
        """Happens after subclass update() for an undoable update by user."""
        oldWidgetVals = {
            attr: copy(getattr(self, attr)) for attr in self._trackedWidgets
        }
        newWidgetVals = {
            attr: (
                getWidgetValue(widget)
                if attr not in self._colorWidgets
                else rgbFromString(widget.text())
            )
            for attr, widget in self._trackedWidgets.items()
        }
        modifiedWidgets = {
            attr: val
            for attr, val in newWidgetVals.items()
            if val != oldWidgetVals[attr]
        }
        if modifiedWidgets:
            action = ComponentUpdate(self, oldWidgetVals, modifiedWidgets)
            self.parent.undoStack.push(action)

    def _autoUpdate(self):
        """Happens after subclass update() for an internal component update."""
        newWidgetVals = {
            attr: getWidgetValue(widget)
            for attr, widget in self._trackedWidgets.items()
        }
        self.setAttrs(newWidgetVals)
        self._sendUpdateSignal()

    def setAttrs(self, attrDict):
        """
        Sets attrs (linked to trackedWidgets) in this component to
        the values in the attrDict. Mutates certain widget values if needed
        """
        for attr, val in attrDict.items():
            if attr in self._colorWidgets:
                # Color Widgets must have a tuple & have a button to update
                if type(val) is tuple:
                    rgbTuple = val
                else:
                    rgbTuple = rgbFromString(val)
                btnStyle = (
                    "QPushButton { background-color : %s; outline: none; }"
                    % QColor(*rgbTuple).name()
                )
                self._colorWidgets[attr].setStyleSheet(btnStyle)
                setattr(self, attr, rgbTuple)

            else:
                # Normal tracked widget
                setattr(self, attr, val)
            log.verbose("Setting %s self.%s to %s" % (self.__class__.name, attr, val))

    def setWidgetValues(self, attrDict):
        """
        Sets widgets defined by keys in trackedWidgets in this preset to
        the values in the attrDict.
        """
        affectedWidgets = [self._trackedWidgets[attr] for attr in attrDict]
        with blockSignals(affectedWidgets):
            for attr, val in attrDict.items():
                widget = self._trackedWidgets[attr]
                if attr in self._colorWidgets:
                    val = "%s,%s,%s" % val
                setWidgetValue(widget, val)

    def _sendUpdateSignal(self):
        if not self.core.openingProject:
            self.parent.drawPreview()
            saveValueStore = self.savePreset()
            saveValueStore["preset"] = self.currentPreset
            self.modified.emit(self.compPos, saveValueStore)

    def trackWidgets(self, trackDict, **kwargs):
        """
        Name widgets to track in update(), savePreset(), loadPreset(), and
        command(). Requires a dict of attr names as keys, widgets as values

        Optional args:
            'presetNames': preset variable names to replace attr names
            'commandArgs': arg keywords that differ from attr names
            'colorWidgets': identify attr as RGB tuple & update button CSS
            'relativeWidgets': change value proportionally to resolution

        NOTE: Any kwarg key set to None will selectively disable tracking.
        """
        self._trackedWidgets = trackDict
        for kwarg in kwargs:
            try:
                if kwarg in (
                    "presetNames",
                    "commandArgs",
                    "colorWidgets",
                    "relativeWidgets",
                ):
                    setattr(self, "_{}".format(kwarg), kwargs[kwarg])
                else:
                    raise ComponentError(self, "Nonsensical keywords to trackWidgets.")
            except ComponentError:
                continue

            if kwarg == "colorWidgets":

                def makeColorFunc(attr):
                    def pickColor_():
                        self.mergeUndo = False
                        self.pickColor(
                            self._trackedWidgets[attr],
                            self._colorWidgets[attr],
                        )
                        self.mergeUndo = True

                    return pickColor_

                self._colorFuncs = {attr: makeColorFunc(attr) for attr in kwargs[kwarg]}
                for attr, func in self._colorFuncs.items():
                    colorText = self._trackedWidgets[attr].text()
                    if colorText == "":
                        rndColor = randomColor()
                        self._trackedWidgets[attr].setText(str(rndColor)[1:-1])
                    self._colorWidgets[attr].clicked.connect(func)
                    self._colorWidgets[attr].setStyleSheet(
                        "QPushButton {"
                        "background-color : %s; outline: none; }"
                        % QColor(
                            *rgbFromString(colorText) if colorText else rndColor
                        ).name()
                    )

            if kwarg == "relativeWidgets":
                # store maximum values of spinBoxes to be scaled appropriately
                for attr in kwargs[kwarg]:
                    self._relativeMaximums[attr] = self._trackedWidgets[attr].maximum()
                    self.updateRelativeWidgetMaximum(attr)
                    setattr(self, attr, getWidgetValue(self._trackedWidgets[attr]))

        self._preUpdate()
        self._autoUpdate()

    def pickColor(self, textWidget, button):
        """Use color picker to get color input from the user."""
        dialog = QtWidgets.QColorDialog()
        # TODO alpha channel is not actually shown in most color picker widgets?
        dialog.setOption(
            QtWidgets.QColorDialog.ColorDialogOption.ShowAlphaChannel, True
        )
        color = dialog.getColor()
        if color.isValid():
            RGBstring = "%s,%s,%s" % (
                str(color.red()),
                str(color.green()),
                str(color.blue()),
            )
            btnStyle = (
                "QPushButton{background-color: %s; outline: none;}" % color.name()
            )
            textWidget.setText(RGBstring)
            button.setStyleSheet(btnStyle)

    def lockProperties(self, propList):
        self._lockedProperties = propList

    def lockError(self, msg):
        self._lockedError = msg

    def lockSize(self, w, h):
        self._lockedSize = (w, h)

    def unlockProperties(self):
        self._lockedProperties = None

    def unlockError(self):
        self._lockedError = None

    def unlockSize(self):
        self._lockedSize = None

    def loadUi(self, filename):
        """Load a Qt Designer ui file to use for this component's widget"""
        return uic.loadUi(os.path.join(self.core.componentsPath, filename))

    @property
    def width(self):
        if self._lockedSize is None:
            return int(self.settings.value("outputWidth"))
        else:
            return self._lockedSize[0]

    @property
    def height(self):
        if self._lockedSize is None:
            return int(self.settings.value("outputHeight"))
        else:
            return self._lockedSize[1]

    def cancel(self):
        """Stop any lengthy process in response to this variable."""
        self.canceled = True

    def reset(self):
        self.canceled = False
        self.unlockProperties()
        self.unlockError()

    def relativeWidgetAxis(func):
        def relativeWidgetAxis(self, attr, *args, **kwargs):
            hasVerticalWords = (
                lambda attr: "height" in attr.lower()
                or "ypos" in attr.lower()
                or attr == "y"
            )
            if "axis" not in kwargs:
                axis = self.width
                if hasVerticalWords(attr):
                    axis = self.height
                kwargs["axis"] = axis
            if "axis" in kwargs and type(kwargs["axis"]) is tuple:
                axis = kwargs["axis"][0]
                if hasVerticalWords(attr):
                    axis = kwargs["axis"][1]
                kwargs["axis"] = axis
            return func(self, attr, *args, **kwargs)

        return relativeWidgetAxis

    @relativeWidgetAxis
    def pixelValForAttr(self, attr, val=None, **kwargs):
        if val is None:
            val = self._relativeValues[attr]
        if val > 50.0:
            log.warning(
                "%s #%s attempted to set %s to dangerously high number %s",
                self.__class__.name,
                self.compPos,
                attr,
                val,
            )
            val = 50.0
        result = math.ceil(kwargs["axis"] * val)
        log.verbose(
            "Converting %s: f%s to px%s using axis %s",
            attr,
            val,
            result,
            kwargs["axis"],
        )
        return result

    @relativeWidgetAxis
    def floatValForAttr(self, attr, val=None, **kwargs):
        if val is None:
            val = self._trackedWidgets[attr].value()
        return val / kwargs["axis"]

    def setRelativeWidget(self, attr, floatVal):
        """Set a relative widget using a float"""
        pixelVal = self.pixelValForAttr(attr, floatVal)
        with blockSignals(self._trackedWidgets[attr]):
            self._trackedWidgets[attr].setValue(pixelVal)
        self.update(auto=True)

    def getOldAttr(self, attr):
        """
        Returns previous state of this attr. Used to determine whether
        a relative widget must be updated. Required because undoing/redoing
        can make determining the 'previous' value tricky.
        """
        if self.oldAttrs is not None:
            return self.oldAttrs[attr]
        else:
            try:
                return getattr(self, attr)
            except AttributeError:
                log.error("Using visible values instead of oldAttrs")
                return self._trackedWidgets[attr].value()

    def updateRelativeWidget(self, attr):
        """Called by _preUpdate() for each relativeWidget before each update"""
        oldUserValue = self.getOldAttr(attr)
        newUserValue = self._trackedWidgets[attr].value()
        newRelativeVal = self.floatValForAttr(attr, newUserValue)

        if attr in self._relativeValues:
            oldRelativeVal = self._relativeValues[attr]
            if oldUserValue == newUserValue and oldRelativeVal != newRelativeVal:
                # Float changed without pixel value changing, which
                # means the pixel value needs to be updated
                # TODO QDoubleSpinBox doesn't work with relativeWidgets because of this
                log.debug(
                    "Updating %s #%s's relative widget: %s",
                    self.__class__.name,
                    self.compPos,
                    attr,
                )
                with blockSignals(self._trackedWidgets[attr]):
                    self.updateRelativeWidgetMaximum(attr)
                    pixelVal = self.pixelValForAttr(attr, oldRelativeVal)
                    self._trackedWidgets[attr].setValue(pixelVal)

        if attr not in self._relativeValues or oldUserValue != newUserValue:
            self._relativeValues[attr] = newRelativeVal

    def updateRelativeWidgetMaximum(self, attr):
        maxRes = int(self.core.resolutions[0].split("x")[0])
        newMaximumValue = self.width * (self._relativeMaximums[attr] / maxRes)
        self._trackedWidgets[attr].setMaximum(int(newMaximumValue))



from avp.command import Command
from pytestqt import qtbot
from pytest import fixture
from . import imageDataSum, command


@fixture
def coreWithLifeComp(qtbot, command):
    """Fixture providing a Command object with Waveform component added"""
    command.settings.setValue("outputHeight", 1080)
    command.settings.setValue("outputWidth", 1920)
    command.core.insertComponent(
        0, command.core.moduleIndexFor("Conway's Game of Life"), command
    )
    yield command.core


def test_comp_life_previewRender(coreWithLifeComp):
    comp = coreWithLifeComp.selectedComponents[0]
    comp.page.lineEdit_color.setText("111,111,111")
    image = comp.previewRender()
    assert imageDataSum(image) == 339785512



import time
import sys
import logging

from ..toolkit import formatTraceback


log = logging.getLogger("AVP.ComponentHandler")


class ComponentError(RuntimeError):
    """Gives the MainWindow a traceback to display, and cancels the export."""

    prevErrors = []
    lastTime = time.time()

    def __init__(self, caller, name, msg=None):
        if msg is None and sys.exc_info()[0] is not None:
            msg = str(sys.exc_info()[1])
        else:
            msg = "Unknown error."
        log.error("ComponentError by %s's %s: %s" % (caller.name, name, msg))

        # Don't create multiple windows for quickly repeated messages
        if len(ComponentError.prevErrors) > 1:
            ComponentError.prevErrors.pop()
        ComponentError.prevErrors.insert(0, name)
        curTime = time.time()
        if (
            name in ComponentError.prevErrors[1:]
            and curTime - ComponentError.lastTime < 1.0
        ):
            return
        ComponentError.lastTime = time.time()

        if sys.exc_info()[0] is not None:
            string = "%s component (#%s): %s encountered %s %s: %s" % (
                caller.__class__.name,
                str(caller.compPos),
                name,
                (
                    "an"
                    if any(
                        [
                            sys.exc_info()[0].__name__.startswith(vowel)
                            for vowel in ("A", "I", "U", "O", "E")
                        ]
                    )
                    else "a"
                ),
                sys.exc_info()[0].__name__,
                str(sys.exc_info()[1]),
            )
            detail = formatTraceback(sys.exc_info()[2])
        else:
            string = name
            detail = "Attributes:\n%s" % (
                "\n".join([m for m in dir(caller) if not m.startswith("_")])
            )

        super().__init__(string)
        caller.lockError(string)
        caller._error.emit(string, detail)



from PyQt6 import QtGui
import logging

from ..libcomponent import BaseComponent
from ..toolkit.frame import BlankFrame, FloodFrame, FramePainter


log = logging.getLogger("AVP.Components.Color")


class Component(BaseComponent):
    name = "Color"
    version = "1.0.0"

    def widget(self, *args):
        self.x = 0
        self.y = 0
        super().widget(*args)

        # disable color #2 until non-default 'fill' option gets changed
        self.page.lineEdit_color2.setDisabled(True)
        self.page.pushButton_color2.setDisabled(True)
        self.page.spinBox_width.setValue(int(self.settings.value("outputWidth")))
        self.page.spinBox_height.setValue(int(self.settings.value("outputHeight")))

        self.fillLabels = [
            "Solid",
            "Linear Gradient",
            "Radial Gradient",
        ]
        for label in self.fillLabels:
            self.page.comboBox_fill.addItem(label)
        self.page.comboBox_fill.setCurrentIndex(0)

        self.trackWidgets(
            {
                "x": self.page.spinBox_x,
                "y": self.page.spinBox_y,
                "sizeWidth": self.page.spinBox_width,
                "sizeHeight": self.page.spinBox_height,
                "trans": self.page.checkBox_trans,
                "spread": self.page.comboBox_spread,
                "stretch": self.page.checkBox_stretch,
                "RG_start": self.page.spinBox_radialGradient_start,
                "LG_start": self.page.spinBox_linearGradient_start,
                "RG_end": self.page.spinBox_radialGradient_end,
                "LG_end": self.page.spinBox_linearGradient_end,
                "RG_centre": self.page.spinBox_radialGradient_spread,
                "fillType": self.page.comboBox_fill,
                "color1": self.page.lineEdit_color1,
                "color2": self.page.lineEdit_color2,
            },
            presetNames={
                "sizeWidth": "width",
                "sizeHeight": "height",
            },
            colorWidgets={
                "color1": self.page.pushButton_color1,
                "color2": self.page.pushButton_color2,
            },
            relativeWidgets=[
                "x",
                "y",
                "sizeWidth",
                "sizeHeight",
                "LG_start",
                "LG_end",
                "RG_start",
                "RG_end",
                "RG_centre",
            ],
        )

    def update(self):
        fillType = self.page.comboBox_fill.currentIndex()
        if fillType == 0:
            self.page.lineEdit_color2.setEnabled(False)
            self.page.pushButton_color2.setEnabled(False)
            self.page.checkBox_trans.setEnabled(False)
            self.page.checkBox_stretch.setEnabled(False)
            self.page.comboBox_spread.setEnabled(False)
        else:
            self.page.lineEdit_color2.setEnabled(True)
            self.page.pushButton_color2.setEnabled(True)
            self.page.checkBox_trans.setEnabled(True)
            self.page.checkBox_stretch.setEnabled(True)
            self.page.comboBox_spread.setEnabled(True)
        if self.page.checkBox_trans.isChecked():
            self.page.lineEdit_color2.setEnabled(False)
            self.page.pushButton_color2.setEnabled(False)
        self.page.fillWidget.setCurrentIndex(fillType)

    def previewRender(self):
        return self.drawFrame(self.width, self.height)

    def properties(self):
        return ["static"]

    def frameRender(self, frameNo):
        log.debug("Color component is drawing frame #%s", frameNo)
        return self.drawFrame(self.width, self.height)

    def drawFrame(self, width, height):
        r, g, b = self.color1
        shapeSize = (self.sizeWidth, self.sizeHeight)
        # in default state, skip all this logic and return a plain fill
        if (
            self.fillType == 0
            and shapeSize == (width, height)
            and self.x == 0
            and self.y == 0
        ):
            return FloodFrame(width, height, (r, g, b, 255))

        # Return a solid image at x, y
        if self.fillType == 0:
            frame = BlankFrame(width, height)
            image = FloodFrame(self.sizeWidth, self.sizeHeight, (r, g, b, 255))
            frame.paste(image, box=(self.x, self.y))
            return frame

        # Now fills that require using Qt...
        elif self.fillType > 0:
            image = FramePainter(width, height)

            if self.stretch:
                w = width
                h = height
            else:
                w = self.sizeWidth
                h = self.sizeWidth

        if self.fillType == 1:  # Linear Gradient
            brush = QtGui.QLinearGradient(
                float(self.LG_start),
                float(self.LG_start),
                float(self.LG_end + width / 3),
                float(self.LG_end),
            )

        elif self.fillType == 2:  # Radial Gradient
            brush = QtGui.QRadialGradient(
                float(self.RG_start),
                float(self.RG_end),
                float(w),
                float(h),
                float(self.RG_centre),
            )
        spread = QtGui.QGradient.Spread.PadSpread
        if self.spread == 1:
            spread = QtGui.QGradient.Spread.ReflectSpread
        elif self.spread == 2:
            spread = QtGui.QGradient.Spread.RepeatSpread
        brush.setSpread(spread)
        brush.setColorAt(0.0, QtGui.QColor(*self.color1))
        if self.trans:
            brush.setColorAt(1.0, QtGui.QColor(0, 0, 0, 0))
        elif self.fillType == 1 and self.stretch:
            brush.setColorAt(0.2, QtGui.QColor(*self.color2))
        else:
            brush.setColorAt(1.0, QtGui.QColor(*self.color2))
        image.setBrush(brush)
        image.drawRect(self.x, self.y, self.sizeWidth, self.sizeHeight)

        return image.finalize()

    def commandHelp(self):
        print("Specify a color:\n    color=255,255,255")

    def command(self, arg):
        if "=" in arg:
            key, arg = arg.split("=", 1)
            if key == "color":
                self.page.lineEdit_color1.setText(arg)
                return
        super().command(arg)



import sys
import os
import logging


__version__ = "2.2.4"


class Logger(logging.getLoggerClass()):
    """
    Custom Logger class to handle custom VERBOSE log level.
    Levels used in this program are as follows:
        VERBOSE     Annoyingly frequent debug messages (e.g, in loops)
        DEBUG       Ordinary debug information
        INFO        Expected events that are expensive or irreversible
        WARNING     A non-fatal error or suspicious behaviour
        ERROR       Any error that would interrupt the user
        CRITICAL    Things that really shouldn't happen at all
    """

    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)
        logging.addLevelName(5, "VERBOSE")

    def verbose(self, msg, *args, **kwargs):
        if self.isEnabledFor(5):
            self._log(5, msg, args, **kwargs)


logging.setLoggerClass(Logger)
logging.VERBOSE = 5


if getattr(sys, "frozen", False):
    # frozen
    wd = os.path.dirname(sys.executable)
else:
    # unfrozen
    wd = os.path.dirname(os.path.realpath(__file__))



import os
from avp.core import Core
from . import getTestDataPath, settings


def test_component_names(settings):
    core = Core()
    assert core.compNames == [
        "Classic Visualizer",
        "Color",
        "Conway's Game of Life",
        "Image",
        "Sound",
        "Spectrum",
        "Title Text",
        "Video",
        "Waveform",
    ]


def test_moduleindex(settings):
    core = Core()
    assert core.moduleIndexFor("Classic Visualizer") == 0


def test_configPath_default():
    configPath = Core.getConfigPath(None)
    assert os.path.basename(configPath) == "audio-visualizer"


def test_configPath_nonstandard():
    assert Core.getConfigPath(getTestDataPath("config")) == getTestDataPath("config")



"""Tests of `actions.py` - MainWindow component list manipulation via undoable actions"""

from PyQt6 import QtCore
import os
from pytest import fixture
from pytestqt import qtbot
from . import getTestDataPath, window


def test_mainwindow_addComponent(qtbot, window):
    window.compMenu.actions()[0].trigger()
    assert len(window.core.selectedComponents) == 1


def test_mainwindow_removeComponent(qtbot, window):
    window.compMenu.actions()[0].trigger()  # add component
    window.pushButton_removeComponent.click()  # remove it
    assert len(window.core.selectedComponents) == 0


def test_mainwindow_moveComponent(qtbot, window):
    # add first two components from menu
    window.compMenu.actions()[0].trigger()
    window.compMenu.actions()[1].trigger()
    comp0 = window.core.selectedComponents[0].ui
    window.pushButton_listMoveDown.click()
    # check if 0 is now 1
    assert window.core.selectedComponents[1].ui == comp0


def test_mainwindow_addComponent_undo(qtbot, window):
    window.compMenu.actions()[0].trigger()
    window.undoStack.undo()
    assert len(window.core.selectedComponents) == 0


def test_mainwindow_removeComponent_undo(qtbot, window):
    window.compMenu.actions()[0].trigger()  # add component
    window.pushButton_removeComponent.click()  # remove it
    window.undoStack.undo()
    assert len(window.core.selectedComponents) == 1


def test_mainwindow_moveComponent_undo(qtbot, window):
    # add first two components from menu
    window.compMenu.actions()[0].trigger()
    window.compMenu.actions()[1].trigger()
    comp0 = window.core.selectedComponents[0].ui
    window.pushButton_listMoveDown.click()
    window.undoStack.undo()
    # check if 0 is still 0 after undo
    assert window.core.selectedComponents[1].ui != comp0



from PyQt6.QtGui import QFont
from PyQt6 import QtGui, QtCore
import logging

from ..libcomponent import BaseComponent
from ..toolkit.frame import FramePainter, addShadow

log = logging.getLogger("AVP.Components.Text")


class Component(BaseComponent):
    name = "Title Text"
    version = "1.0.1"

    def widget(self, *args):
        super().widget(*args)
        self.title = "Text"
        self.alignment = 1
        self.titleFont = QFont()
        self.fontSize = self.height / 13.5

        self.page.comboBox_textAlign.addItem("Left")
        self.page.comboBox_textAlign.addItem("Middle")
        self.page.comboBox_textAlign.addItem("Right")
        self.page.comboBox_textAlign.setCurrentIndex(int(self.alignment))
        self.page.spinBox_fontSize.setValue(int(self.fontSize))
        self.page.pushButton_center.clicked.connect(self.centerXY)

        self.page.fontComboBox_titleFont.currentFontChanged.connect(
            self._sendUpdateSignal
        )
        # The QFontComboBox must be connected directly to the Qt Signal
        # which triggers the preview to update.
        # This unfortunately makes changing the font into a non-undoable action.
        # Fix requires updating ComponentAction to handle fonts

        self.trackWidgets(
            {
                "textColor": self.page.lineEdit_textColor,
                "title": self.page.lineEdit_title,
                "alignment": self.page.comboBox_textAlign,
                "fontSize": self.page.spinBox_fontSize,
                "xPosition": self.page.spinBox_xTextAlign,
                "yPosition": self.page.spinBox_yTextAlign,
                "fontStyle": self.page.comboBox_fontStyle,
                "stroke": self.page.spinBox_stroke,
                "strokeColor": self.page.lineEdit_strokeColor,
                "shadow": self.page.checkBox_shadow,
                "shadX": self.page.spinBox_shadX,
                "shadY": self.page.spinBox_shadY,
                "shadBlur": self.page.spinBox_shadBlur,
            },
            colorWidgets={
                "textColor": self.page.pushButton_textColor,
                "strokeColor": self.page.pushButton_strokeColor,
            },
            relativeWidgets=[
                "xPosition",
                "yPosition",
                "fontSize",
                "stroke",
                "shadX",
                "shadY",
                "shadBlur",
            ],
        )
        self.centerXY()

    def update(self):
        self.titleFont = self.page.fontComboBox_titleFont.currentFont()
        if self.page.checkBox_shadow.isChecked():
            self.page.label_shadX.setHidden(False)
            self.page.spinBox_shadX.setHidden(False)
            self.page.spinBox_shadY.setHidden(False)
            self.page.label_shadBlur.setHidden(False)
            self.page.spinBox_shadBlur.setHidden(False)
        else:
            self.page.label_shadX.setHidden(True)
            self.page.spinBox_shadX.setHidden(True)
            self.page.spinBox_shadY.setHidden(True)
            self.page.label_shadBlur.setHidden(True)
            self.page.spinBox_shadBlur.setHidden(True)

    def centerXY(self):
        self.setRelativeWidget("xPosition", 0.5)
        self.setRelativeWidget("yPosition", 0.521)

    def getXY(self):
        """Returns true x, y after considering alignment settings"""
        fm = QtGui.QFontMetrics(self.titleFont)
        text_width = fm.boundingRect(self.title).width()
        x = self.pixelValForAttr("xPosition")

        if self.alignment == 1:  # Middle
            offset = int(text_width / 2)
        elif self.alignment == 2:  # Right
            offset = text_width
        else:
            raise ValueError(f"Alignment value {self.alignment} unknown")

        x -= offset

        return x, self.yPosition

    def loadPreset(self, pr, *args):
        super().loadPreset(pr, *args)

        font = QFont()
        font.fromString(pr["titleFont"])
        self.page.fontComboBox_titleFont.setCurrentFont(font)

    def savePreset(self):
        saveValueStore = super().savePreset()
        saveValueStore["titleFont"] = self.titleFont.toString()
        return saveValueStore

    def previewRender(self):
        return self.addText(self.width, self.height)

    def properties(self):
        props = ["static"]
        if not self.title:
            props.append("error")
        return props

    def error(self):
        return "No text provided."

    def frameRender(self, frameNo):
        return self.addText(self.width, self.height)

    def addText(self, width, height):
        font = self.titleFont
        font.setPixelSize(self.fontSize)
        font.setStyle(QFont.Style.StyleNormal)
        font.setWeight(QFont.Weight.Normal)
        font.setCapitalization(QFont.Capitalization.MixedCase)
        if self.fontStyle == 1:
            font.setWeight(QFont.Weight.DemiBold)
        if self.fontStyle == 2:
            font.setWeight(QFont.Weight.Bold)
        elif self.fontStyle == 3:
            font.setStyle(QFont.Style.StyleItalic)
        elif self.fontStyle == 4:
            font.setWeight(QFont.Weight.Bold)
            font.setStyle(QFont.Style.StyleItalic)
        elif self.fontStyle == 5:
            font.setStyle(QFont.Style.StyleOblique)
        elif self.fontStyle == 6:
            font.setCapitalization(QFont.Capitalization.SmallCaps)

        image = FramePainter(width, height)
        x, y = self.getXY()
        log.debug("Text position translates to %s, %s", x, y)
        if self.stroke > 0:
            outliner = QtGui.QPainterPathStroker()
            outliner.setWidth(self.stroke)
            path = QtGui.QPainterPath()
            if self.fontStyle == 6:
                # PathStroker ignores smallcaps so we need this weird hack
                path.addText(x, y, font, self.title[0])
                fm = QtGui.QFontMetrics(font)
                newX = x + fm.boundingRect(self.title[0]).width()
                strokeFont = self.page.fontComboBox_titleFont.currentFont()
                strokeFont.setCapitalization(QFont.Capitalization.SmallCaps)
                strokeFont.setPixelSize(int((self.fontSize / 7) * 5))
                strokeFont.setLetterSpacing(QFont.SpacingType.PercentageSpacing, 139)
                path.addText(newX, y, strokeFont, self.title[1:])
            else:
                path.addText(x, y, font, self.title)
            path = outliner.createStroke(path)
            image.setPen(QtCore.Qt.PenStyle.NoPen)
            image.setBrush(QtGui.QColor(*self.strokeColor))
            image.drawPath(path)

        image.setFont(font)
        image.setPen(self.textColor)
        image.drawText(x, y, self.title)

        # turn QImage into Pillow frame
        frame = image.finalize()
        if self.shadow:
            frame = addShadow(frame, self.shadBlur / 10, self.shadX, self.shadY)

        return frame

    def commandHelp(self):
        print("Enter a string to use as centred white text:")
        print('    "title=User Error"')
        print("Specify a text color:\n    color=255,255,255")
        print("Set custom x, y position:\n    x=500 y=500")

    def command(self, arg):
        if "=" in arg:
            key, arg = arg.split("=", 1)
            if key == "color":
                self.page.lineEdit_textColor.setText(arg)
                return
            elif key == "size":
                self.page.spinBox_fontSize.setValue(int(arg))
                return
            elif key == "x":
                self.page.spinBox_xTextAlign.setValue(int(arg))
                return
            elif key == "y":
                self.page.spinBox_yTextAlign.setValue(int(arg))
                return
            elif key == "title":
                self.page.lineEdit_title.setText(arg)
                return
        super().command(arg)



import sys
import os
import tempfile
from . import command, getTestDataPath, MockSignal
from pytestqt import qtbot


def test_commandline_classic_export(qtbot, command):
    """Run Qt event loop and create a video in the system /tmp or /temp"""
    soundFile = getTestDataPath("inputfiles/test.ogg")
    outputDir = tempfile.mkdtemp(prefix="avp-export-")
    outputFilename = os.path.join(outputDir, "output.mp4")
    sys.argv = [
        "",
        "-c",
        "0",
        "classic",
        "color=255,255,255",
        "-i",
        soundFile,
        "-o",
        outputFilename,
    ]
    command.parseArgs()
    # Command object now has a video_thread Worker which is exporting the video

    with qtbot.waitSignal(command.worker.videoCreated, timeout=10000):
        """
        Wait until videoCreated is emitted by the video_thread Worker
        or until 10 second timeout has passed
        """
        print(f"Test Video created at {outputFilename}")

    assert os.path.exists(outputFilename)
    # output video should be at least 200kb
    assert os.path.getsize(outputFilename) > 200000



import os
from avp.command import Command
from pytestqt import qtbot
from pytest import fixture
from . import audioData, command, MockSignal, imageDataSum, getTestDataPath


sampleSize = 1470  # 44100 / 30 = 1470
testFile = "inputfiles/test.jpg"


@fixture
def coreWithImageComp(qtbot, command):
    """Fixture providing a Command object with Image component added"""
    command.settings.setValue("outputHeight", 1080)
    command.settings.setValue("outputWidth", 1920)
    command.core.insertComponent(0, command.core.moduleIndexFor("Image"), command)
    yield command.core


def test_comp_image_set_path(coreWithImageComp):
    "Set imagePath of Image component"
    comp = coreWithImageComp.selectedComponents[0]
    comp.imagePath = getTestDataPath(testFile)
    image = comp.previewRender()
    assert imageDataSum(image) == 463711601


def test_comp_image_scale_50_1080p(coreWithImageComp):
    """Image component stretches image to 50% at 1080p"""
    comp = coreWithImageComp.selectedComponents[0]
    comp.imagePath = getTestDataPath(testFile)
    image = comp.previewRender()
    sum = imageDataSum(image)
    comp.page.spinBox_scale.setValue(50)
    assert imageDataSum(comp.previewRender()) - sum / 4 < 2000


def test_comp_image_scale_50_720p(coreWithImageComp):
    """Image component stretches image to 50% at 720p"""
    comp = coreWithImageComp.selectedComponents[0]
    comp.imagePath = getTestDataPath(testFile)
    comp.page.spinBox_scale.setValue(50)
    image = comp.previewRender()
    sum = imageDataSum(image)
    comp.parent.settings.setValue("outputHeight", 720)
    comp.parent.settings.setValue("outputWidth", 1280)
    newImage = comp.previewRender()
    assert image.width == 1920
    assert newImage.width == 1280
    assert imageDataSum(comp.previewRender()) == sum


def test_comp_image_command_path(coreWithImageComp):
    """Image component accepts commandline argument:
    `path=test.jpg`"""
    imgPath = os.path.realpath(getTestDataPath(testFile))
    comp = coreWithImageComp.selectedComponents[0]
    comp.command(f"path={imgPath}")
    assert comp.imagePath == imgPath



"""Tests of MainWindow undoing certain ComponentActions (changes to component settings)"""

from pytest import fixture
from pytestqt import qtbot
from avp.gui.mainwindow import MainWindow
from . import getTestDataPath, window


def test_undo_classic_visualizer_sensitivity(window, qtbot):
    """Undo Classic Visualizer component sensitivity setting
    should undo multiple merged actions."""
    window.core.insertComponent(
        0, window.core.moduleIndexFor("Classic Visualizer"), window
    )
    comp = window.core.selectedComponents[0]
    comp.imagePath = getTestDataPath("inputfiles/test.jpg")
    for i in range(1, 100):
        comp.page.spinBox_scale.setValue(i)
    assert comp.scale == 99
    window.undoStack.undo()
    assert comp.scale == 20


def test_undo_image_scale(window, qtbot):
    """Undo Image component scale setting should undo multiple merged actions."""
    window.core.insertComponent(0, window.core.moduleIndexFor("Image"), window)
    comp = window.core.selectedComponents[0]
    comp.imagePath = getTestDataPath("inputfiles/test.jpg")
    comp.page.spinBox_scale.setValue(100)
    for i in range(10, 401):
        comp.page.spinBox_scale.setValue(i)
    assert comp.scale == 400
    window.undoStack.undo()
    assert comp.scale == 10
    window.undoStack.undo()
    assert comp.scale == 100


def test_undo_image_resizeMode(window, qtbot):
    window.core.insertComponent(0, window.core.moduleIndexFor("Image"), window)
    comp = window.core.selectedComponents[0]
    comp.page.comboBox_resizeMode.setCurrentIndex(1)
    assert not comp.page.spinBox_scale.isEnabled()
    window.undoStack.undo()
    assert comp.page.spinBox_scale.isEnabled()


def test_undo_title_text_merged(window, qtbot):
    """Undoing title text change should undo all recent changes."""
    window.core.insertComponent(0, window.core.moduleIndexFor("Title Text"), window)
    comp = window.core.selectedComponents[0]
    comp.page.lineEdit_title.setText("avp")
    comp.page.lineEdit_title.setText("test")
    window.undoStack.undo()
    assert comp.title == "Text"


def test_undo_title_text_not_merged(window, qtbot):
    """Undoing title text change should undo up to previous different action"""
    window.core.insertComponent(0, window.core.moduleIndexFor("Title Text"), window)
    comp = window.core.selectedComponents[0]
    comp.page.lineEdit_title.setText("avp")
    comp.page.spinBox_xTextAlign.setValue(0)
    comp.page.lineEdit_title.setText("test")
    window.undoStack.undo()
    assert comp.title == "avp"



from PyQt6.QtGui import QUndoStack
from ..toolkit.common import disableWhenEncoding


class UndoStack(QUndoStack):
    @property
    def encoding(self):
        return self.parent().encoding

    @disableWhenEncoding
    def undo(self, *args, **kwargs):
        super().undo(*args, **kwargs)

    @disableWhenEncoding
    def redo(self, *args, **kwargs):
        super().redo(*args, **kwargs)



"""
Common functions
"""

from PyQt6 import QtWidgets
import string
import random
import sys
import subprocess
import logging
from copy import copy
from collections import OrderedDict


log = logging.getLogger("AVP.Toolkit.Common")


class blockSignals:
    """
    Context manager to temporarily block list of QtWidgets from updating,
    and guarantee restoring the previous state afterwards.
    """

    def __init__(self, widgets):
        if type(widgets) is dict:
            self.widgets = concatDictVals(widgets)
        else:
            self.widgets = widgets if hasattr(widgets, "__iter__") else [widgets]

    def __enter__(self):
        log.verbose(
            "Blocking signals for %s",
            ", ".join([str(w.__class__.__name__) for w in self.widgets]),
        )
        self.oldStates = [w.signalsBlocked() for w in self.widgets]
        for w in self.widgets:
            w.blockSignals(True)

    def __exit__(self, *args):
        log.verbose("Resetting blockSignals to %s", str(bool(sum(self.oldStates))))
        for w, state in zip(self.widgets, self.oldStates):
            w.blockSignals(state)


def concatDictVals(d):
    """Concatenates all values in given dict into one list."""
    key, value = d.popitem()
    d[key] = value
    final = copy(value)
    if type(final) is not list:
        final = [final]
        final.extend([val for val in d.values()])
    else:
        value.extend([item for val in d.values() for item in val])
    return final


def badName(name):
    """Returns whether a name contains non-alphanumeric chars"""
    return any([letter in string.punctuation for letter in name])


def alphabetizeDict(dictionary):
    """Alphabetizes a dict into OrderedDict"""
    return OrderedDict(sorted(dictionary.items(), key=lambda t: t[0]))


def presetToString(dictionary):
    """Returns string repr of a preset"""
    return repr(alphabetizeDict(dictionary))


def presetFromString(string):
    """Turns a string repr of OrderedDict into a regular dict"""
    return dict(eval(string))


def appendUppercase(lst):
    for form, i in zip(lst, range(len(lst))):
        lst.append(form.upper())
    return lst


def pipeWrapper(func):
    """A decorator to insert proper kwargs into Popen objects."""

    def pipeWrapper(commandList, **kwargs):
        if sys.platform == "win32":
            # Stop CMD window from appearing on Windows
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            kwargs["startupinfo"] = startupinfo

        if "bufsize" not in kwargs:
            kwargs["bufsize"] = 10**8
        if "stdin" not in kwargs:
            kwargs["stdin"] = subprocess.DEVNULL
        return func(commandList, **kwargs)

    return pipeWrapper


@pipeWrapper
def checkOutput(commandList, **kwargs):
    return subprocess.check_output(commandList, **kwargs)


def disableWhenEncoding(func):
    def decorator(self, *args, **kwargs):
        if self.encoding:
            return
        else:
            return func(self, *args, **kwargs)

    return decorator


def disableWhenOpeningProject(func):
    def decorator(self, *args, **kwargs):
        if self.core.openingProject:
            return
        else:
            return func(self, *args, **kwargs)

    return decorator


def rgbFromString(string):
    """Turns an RGB string like "255, 255, 255" into a tuple"""
    try:
        tup = tuple([int(i) for i in string.split(",")])
        if len(tup) != 3:
            raise ValueError
        for i in tup:
            if i > 255 or i < 0:
                raise ValueError
        return tup
    except Exception as e:
        log.warning(
            "Could not parse '%s' as a color (encountered %s).",
            string,
            type(e).__name__,
        )
        return (255, 255, 255)


def formatTraceback(tb=None):
    import traceback

    if tb is None:
        import sys

        tb = sys.exc_info()[2]
    return "Traceback:\n%s" % "\n".join(traceback.format_tb(tb))


def connectWidget(widget, func):
    unsupportedWidgets = ["QtWidgets.QFontComboBox"]
    if type(widget) == QtWidgets.QLineEdit:
        widget.textChanged.connect(func)
    elif type(widget) == QtWidgets.QSpinBox or type(widget) == QtWidgets.QDoubleSpinBox:
        widget.valueChanged.connect(func)
    elif type(widget) == QtWidgets.QCheckBox:
        widget.stateChanged.connect(func)
    elif type(widget) == QtWidgets.QComboBox:
        widget.currentIndexChanged.connect(func)
    elif type(widget) in unsupportedWidgets:
        log.info(
            "Could not connect %s using connectWidget()", str(widget.__class__.__name__)
        )
    else:
        log.warning("Failed to connect %s ", str(widget.__class__.__name__))
        return False
    return True


def setWidgetValue(widget, val):
    """Generic setValue method for use with any typical QtWidget"""
    log.verbose("Setting %s to %s" % (str(widget.__class__.__name__), val))
    if type(widget) == QtWidgets.QLineEdit:
        widget.setText(val)
    elif type(widget) == QtWidgets.QSpinBox or type(widget) == QtWidgets.QDoubleSpinBox:
        widget.setValue(val)
    elif type(widget) == QtWidgets.QCheckBox:
        widget.setChecked(val)
    elif type(widget) == QtWidgets.QComboBox:
        widget.setCurrentIndex(val)
    else:
        log.warning("Failed to set %s ", str(widget.__class__.__name__))
        return False
    return True


def getWidgetValue(widget):
    if type(widget) == QtWidgets.QLineEdit:
        return widget.text()
    elif type(widget) == QtWidgets.QSpinBox or type(widget) == QtWidgets.QDoubleSpinBox:
        return widget.value()
    elif type(widget) == QtWidgets.QCheckBox:
        return widget.isChecked()
    elif type(widget) == QtWidgets.QComboBox:
        return widget.currentIndex()


def randomColor():
    return (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))



import os
import logging
from PyQt6 import QtCore

from .exceptions import ComponentError
from ..toolkit import connectWidget
from ..toolkit.frame import BlankFrame

log = logging.getLogger("AVP.ComponentHandler")


class ComponentMetaclass(type(QtCore.QObject)):
    """
    Checks the validity of each Component class and mutates some attrs.
    E.g., takes only major version from version string & decorates methods
    """

    def initializationWrapper(func):
        def initializationWrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception:
                try:
                    raise ComponentError(self, "initialization process")
                except ComponentError:
                    return

        return initializationWrapper

    def renderWrapper(func):
        def renderWrapper(self, *args, **kwargs):
            try:
                log.verbose(
                    "### %s #%s renders a preview frame ###",
                    self.__class__.name,
                    str(self.compPos),
                )
                return func(self, *args, **kwargs)
            except Exception as e:
                try:
                    if e.__class__.__name__.startswith("Component"):
                        raise
                    else:
                        raise ComponentError(self, "renderer")
                except ComponentError:
                    return BlankFrame()

        return renderWrapper

    def commandWrapper(func):
        """Intercepts the command() method to check for global args"""

        def commandWrapper(self, arg):
            if arg.startswith("preset="):
                _, preset = arg.split("=", 1)
                path = os.path.join(self.core.getPresetDir(self), preset)
                if not os.path.exists(path):
                    print('Couldn\'t locate preset "%s"' % preset)
                    quit(1)
                else:
                    print('Opening "%s" preset on layer %s' % (preset, self.compPos))
                    self.core.openPreset(path, self.compPos, preset)
                    # Don't call the component's command() method
                    return
            else:
                return func(self, arg)

        return commandWrapper

    def propertiesWrapper(func):
        """Intercepts the usual properties if the properties are locked."""

        def propertiesWrapper(self):
            if self._lockedProperties is not None:
                return self._lockedProperties
            else:
                try:
                    return func(self)
                except Exception:
                    try:
                        raise ComponentError(self, "properties")
                    except ComponentError:
                        return []

        return propertiesWrapper

    def errorWrapper(func):
        """Intercepts the usual error message if it is locked."""

        def errorWrapper(self):
            if self._lockedError is not None:
                return self._lockedError
            else:
                return func(self)

        return errorWrapper

    def loadPresetWrapper(func):
        """Wraps loadPreset to handle the self.openingPreset boolean"""

        class openingPreset:
            def __init__(self, comp):
                self.comp = comp

            def __enter__(self):
                self.comp.openingPreset = True

            def __exit__(self, *args):
                self.comp.openingPreset = False

        def presetWrapper(self, *args):
            with openingPreset(self):
                try:
                    return func(self, *args)
                except Exception:
                    try:
                        raise ComponentError(self, "preset loader")
                    except ComponentError:
                        return

        return presetWrapper

    def updateWrapper(func):
        """
        Calls _preUpdate before every subclass update().
        Afterwards, for non-user updates, calls _autoUpdate().
        For undoable updates triggered by the user, calls _userUpdate()
        """

        class wrap:
            def __init__(self, comp, auto):
                self.comp = comp
                self.auto = auto

            def __enter__(self):
                self.comp._preUpdate()

            def __exit__(self, *args):
                if (
                    self.auto
                    or self.comp.openingPreset
                    or not hasattr(self.comp.parent, "undoStack")
                ):
                    log.verbose("Automatic update")
                    self.comp._autoUpdate()
                else:
                    log.verbose("User update")
                    self.comp._userUpdate()

        def updateWrapper(self, **kwargs):
            auto = kwargs["auto"] if "auto" in kwargs else False
            with wrap(self, auto):
                try:
                    return func(self)
                except Exception:
                    try:
                        raise ComponentError(self, "update method")
                    except ComponentError:
                        return

        return updateWrapper

    def widgetWrapper(func):
        """Connects all widgets to update method after the subclass's method"""

        class wrap:
            def __init__(self, comp):
                self.comp = comp

            def __enter__(self):
                pass

            def __exit__(self, *args):
                for widgetList in self.comp._allWidgets.values():
                    for widget in widgetList:
                        log.verbose("Connecting %s", str(widget.__class__.__name__))
                        connectWidget(widget, self.comp.update)

        def widgetWrapper(self, *args, **kwargs):
            auto = kwargs["auto"] if "auto" in kwargs else False
            with wrap(self):
                try:
                    return func(self, *args, **kwargs)
                except Exception:
                    try:
                        raise ComponentError(self, "widget creation")
                    except ComponentError:
                        return

        return widgetWrapper

    def __new__(cls, name, parents, attrs):
        if "ui" not in attrs:
            # Use module name as ui filename by default
            attrs["ui"] = (
                "%s.ui" % os.path.splitext(attrs["__module__"].split(".")[-1])[0]
            )

        decorate = (
            "names",  # Class methods
            "error",
            "audio",
            "properties",  # Properties
            "preFrameRender",
            "previewRender",
            "loadPreset",
            "command",
            "update",
            "widget",
        )

        # Auto-decorate methods
        for key in decorate:
            if key not in attrs:
                continue
            if key in ("names"):
                attrs[key] = classmethod(attrs[key])
            elif key in ("audio"):
                attrs[key] = property(attrs[key])
            elif key == "command":
                attrs[key] = cls.commandWrapper(attrs[key])
            elif key == "previewRender":
                attrs[key] = cls.renderWrapper(attrs[key])
            elif key == "preFrameRender":
                attrs[key] = cls.initializationWrapper(attrs[key])
            elif key == "properties":
                attrs[key] = cls.propertiesWrapper(attrs[key])
            elif key == "error":
                attrs[key] = cls.errorWrapper(attrs[key])
            elif key == "loadPreset":
                attrs[key] = cls.loadPresetWrapper(attrs[key])
            elif key == "update":
                attrs[key] = cls.updateWrapper(attrs[key])
            elif key == "widget" and parents[0] != QtCore.QObject:
                attrs[key] = cls.widgetWrapper(attrs[key])

        # Turn version string into a number
        try:
            if "version" not in attrs:
                log.error(
                    "No version attribute in %s. Defaulting to 1",
                    attrs["name"],
                )
                attrs["version"] = 1
            else:
                attrs["version"] = int(attrs["version"].split(".")[0])
        except ValueError:
            log.critical(
                "%s component has an invalid version string:\n%s",
                attrs["name"],
                str(attrs["version"]),
            )
        except KeyError:
            log.critical("%s component has no version string.", attrs["name"])
        else:
            return super().__new__(cls, name, parents, attrs)
        quit(1)



"""
Tools for using ffmpeg
"""

import numpy
import sys
import os
import subprocess
import threading
import signal
from queue import PriorityQueue
import logging

from ..core import Core
from .common import checkOutput, pipeWrapper


log = logging.getLogger("AVP.Toolkit.Ffmpeg")


class FfmpegVideo:
    """Opens an input pipe to ffmpeg and stores a buffer of raw video frames."""

    # error from the thread used to fill the buffer
    threadError = None

    def __init__(self, **kwargs):
        mandatoryArgs = [
            "inputPath",
            "filter_",
            "width",
            "height",
            "frameRate",  # frames per second
            "chunkSize",  # number of bytes in one frame
            "parent",  # mainwindow object
            "component",  # component object
        ]
        for arg in mandatoryArgs:
            setattr(self, arg, kwargs[arg])

        self.frameNo = -1
        self.currentFrame = "None"
        self.map_ = None

        if "loopVideo" in kwargs and kwargs["loopVideo"]:
            self.loopValue = "-1"
        else:
            self.loopValue = "0"
        if "filter_" in kwargs:
            if kwargs["filter_"][0] != "-filter_complex":
                kwargs["filter_"].insert(0, "-filter_complex")
        else:
            kwargs["filter_"] = None

        self.command = [
            Core.FFMPEG_BIN,
            "-thread_queue_size",
            "512",
            "-r",
            str(self.frameRate),
            "-stream_loop",
            str(self.loopValue),
            "-i",
            self.inputPath,
            "-f",
            "image2pipe",
            "-pix_fmt",
            "rgba",
        ]
        if type(kwargs["filter_"]) is list:
            self.command.extend(kwargs["filter_"])
        self.command.extend(
            [
                "-codec:v",
                "rawvideo",
                "-",
            ]
        )

        self.frameBuffer = PriorityQueue()
        self.frameBuffer.maxsize = self.frameRate
        self.finishedFrames = {}

        self.thread = threading.Thread(
            target=self.fillBuffer, name="FFmpeg Frame-Fetcher"
        )
        self.thread.daemon = True
        self.thread.start()

    def frame(self, num):
        while True:
            if num in self.finishedFrames:
                image = self.finishedFrames.pop(num)
                return image

            i, image = self.frameBuffer.get()
            self.finishedFrames[i] = image
            self.frameBuffer.task_done()

    def fillBuffer(self):
        from ..libcomponent import ComponentError

        if Core.logEnabled:
            logFilename = os.path.join(
                Core.logDir, "render_%s.log" % str(self.component.compPos)
            )
            log.debug("Creating ffmpeg process (log at %s)", logFilename)
            with open(logFilename, "w") as logf:
                logf.write(" ".join(self.command) + "\n\n")
            with open(logFilename, "a") as logf:
                self.pipe = openPipe(
                    self.command,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess.PIPE,
                    stderr=logf,
                    bufsize=10**8,
                )
        else:
            self.pipe = openPipe(
                self.command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                bufsize=10**8,
            )

        while True:
            if self.parent.canceled:
                break
            self.frameNo += 1

            # If we run out of frames, use the last good frame and loop.
            try:
                if len(self.currentFrame) == 0:
                    self.frameBuffer.put((self.frameNo - 1, self.lastFrame))
                    continue
            except AttributeError:
                FfmpegVideo.threadError = ComponentError(
                    self.component,
                    "video",
                    "Video seemed playable but wasn't.",
                )
                break

            try:
                self.currentFrame = self.pipe.stdout.read(self.chunkSize)
            except ValueError as e:
                if str(e) == "PyMemoryView_FromBuffer(): info->buf must not be NULL":
                    log.debug(
                        "Ignored 'info->buf must not be NULL' error from FFmpeg pipe"
                    )
                    return
                else:
                    FfmpegVideo.threadError = ComponentError(self.component, "video")

            if len(self.currentFrame) != 0:
                self.frameBuffer.put((self.frameNo, self.currentFrame))
                self.lastFrame = self.currentFrame


@pipeWrapper
def openPipe(commandList, **kwargs):
    return subprocess.Popen(commandList, **kwargs)


def closePipe(pipe):
    pipe.stdout.close()
    pipe.send_signal(signal.SIGTERM)


def findFfmpeg():
    if sys.platform == "win32":
        bin = "ffmpeg.exe"
    else:
        bin = "ffmpeg"

    if getattr(sys, "frozen", False):
        # The application is frozen
        bin = os.path.join(Core.wd, bin)

    with open(os.devnull, "w") as f:
        try:
            checkOutput([bin, "-version"], stderr=f)
        except (subprocess.CalledProcessError, FileNotFoundError):
            bin = ""

    return bin


def createFfmpegCommand(
    inputFile, outputFile, components, duration=-1, logLevel="info"
):
    """
    Constructs the major ffmpeg command used to export the video
    """
    if duration == -1:
        duration = getAudioDuration(inputFile)
    safeDuration = "{0:.3f}".format(duration - 0.05)  # used by filters
    duration = "{0:.3f}".format(duration + 0.1)  # used by input sources

    # Test if user has libfdk_aac
    encoders = checkOutput("%s -encoders -hide_banner" % Core.FFMPEG_BIN, shell=True)
    encoders = encoders.decode("utf-8")

    acodec = Core.settings.value("outputAudioCodec")

    options = Core.encoderOptions
    containerName = Core.settings.value("outputContainer")
    vcodec = Core.settings.value("outputVideoCodec")
    vbitrate = str(Core.settings.value("outputVideoBitrate")) + "k"
    acodec = Core.settings.value("outputAudioCodec")
    abitrate = str(Core.settings.value("outputAudioBitrate")) + "k"

    for cont in options["containers"]:
        if cont["name"] == containerName:
            container = cont["container"]
            break

    vencoders = options["video-codecs"][vcodec]
    aencoders = options["audio-codecs"][acodec]

    def error():
        nonlocal encoders, encoder
        log.critical(
            "Selected encoder (%s) is not supported by Ffmpeg. The supported encoders are: %s",
            encoder,
            encoders,
        )
        return []

    for encoder in vencoders:
        if encoder in encoders:
            vencoder = encoder
            break
    else:
        return error()

    for encoder in aencoders:
        if encoder in encoders:
            aencoder = encoder
            break
    else:
        return error()

    ffmpegCommand = [
        Core.FFMPEG_BIN,
        "-loglevel",
        logLevel,
        "-thread_queue_size",
        "512",
        "-y",  # overwrite the output file if it already exists.
        # INPUT VIDEO
        "-f",
        "rawvideo",
        "-vcodec",
        "rawvideo",
        "-s",
        f'{Core.settings.value("outputWidth")}x{Core.settings.value("outputHeight")}',
        "-pix_fmt",
        "rgba",
        "-r",
        str(Core.settings.value("outputFrameRate")),
        "-t",
        duration,
        "-an",  # the video input has no sound
        "-i",
        "-",  # the video input comes from a pipe
        # INPUT SOUND
        "-t",
        duration,
        "-i",
        inputFile,
    ]

    extraAudio = [comp.audio for comp in components if "audio" in comp.properties()]
    segment = createAudioFilterCommand(extraAudio, safeDuration)
    ffmpegCommand.extend(segment)
    # Map audio from the filters or the single audio input, and map video from the pipe
    ffmpegCommand.extend(
        [
            "-map",
            "0:v",
            "-map",
            "[a]" if segment else "1:a",
        ]
    )

    ffmpegCommand.extend(
        [
            # OUTPUT
            "-vcodec",
            vencoder,
            "-acodec",
            aencoder,
            "-b:v",
            vbitrate,
            "-b:a",
            abitrate,
            "-pix_fmt",
            Core.settings.value("outputVideoFormat"),
            "-preset",
            Core.settings.value("outputPreset"),
            "-f",
            container,
        ]
    )

    if acodec == "aac":
        ffmpegCommand.append("-strict")
        ffmpegCommand.append("-2")

    ffmpegCommand.append(outputFile)
    return ffmpegCommand


def createAudioFilterCommand(extraAudio, duration):
    """Add extra inputs and any needed filters to the main ffmpeg command."""
    # NOTE: Global filters are currently hard-coded here for debugging use
    globalFilters = 0  # increase to add global filters

    if not extraAudio and not globalFilters:
        return []

    ffmpegCommand = []
    # Add -i options for extra input files
    extraFilters = {}
    for streamNo, params in enumerate(reversed(extraAudio)):
        extraInputFile, params = params
        ffmpegCommand.extend(
            [
                "-t",
                duration,
                # Tell ffmpeg about shorter clips (seemingly not needed)
                #   streamDuration = getAudioDuration(extraInputFile)
                #   if streamDuration and streamDuration > float(safeDuration)
                #   else "{0:.3f}".format(streamDuration),
                "-i",
                extraInputFile,
            ]
        )
        # Construct dataset of extra filters we'll need to add later
        for ffmpegFilter in params:
            if streamNo + 2 not in extraFilters:
                extraFilters[streamNo + 2] = []
            extraFilters[streamNo + 2].append((ffmpegFilter, params[ffmpegFilter]))

    # Start creating avfilters! Popen-style, so don't use semicolons;
    extraFilterCommand = []

    if globalFilters <= 0:
        # Dictionary of last-used tmp labels for a given stream number
        tmpInputs = {streamNo: -1 for streamNo in extraFilters}
    else:
        # Insert blank entries for global filters into extraFilters
        # so the per-stream filters know what input to source later
        for streamNo in range(len(extraAudio), 0, -1):
            if streamNo + 1 not in extraFilters:
                extraFilters[streamNo + 1] = []
        # Also filter the primary audio track
        extraFilters[1] = []
        tmpInputs = {streamNo: globalFilters - 1 for streamNo in extraFilters}

        # Add the global filters!
        # NOTE: list length must = globalFilters, currently hardcoded
        if tmpInputs:
            extraFilterCommand.extend(
                [
                    "[%s:a] ashowinfo [%stmp0]" % (str(streamNo), str(streamNo))
                    for streamNo in tmpInputs
                ]
            )

    # Now add the per-stream filters!
    for streamNo, paramList in extraFilters.items():
        for param in paramList:
            source = (
                "[%s:a]" % str(streamNo)
                if tmpInputs[streamNo] == -1
                else "[%stmp%s]" % (str(streamNo), str(tmpInputs[streamNo]))
            )
            tmpInputs[streamNo] = tmpInputs[streamNo] + 1
            extraFilterCommand.append(
                "%s %s%s [%stmp%s]"
                % (
                    source,
                    param[0],
                    param[1],
                    str(streamNo),
                    str(tmpInputs[streamNo]),
                )
            )

    # Join all the filters together and combine into 1 stream
    extraFilterCommand = "; ".join(extraFilterCommand) + "; " if tmpInputs else ""
    ffmpegCommand.extend(
        [
            "-filter_complex",
            extraFilterCommand
            + "%s amix=inputs=%s:duration=first [a]"
            % (
                "".join(
                    [
                        (
                            "[%stmp%s]" % (str(i), tmpInputs[i])
                            if i in extraFilters
                            else "[%s:a]" % str(i)
                        )
                        for i in range(1, len(extraAudio) + 2)
                    ]
                ),
                str(len(extraAudio) + 1),
            ),
        ]
    )
    return ffmpegCommand


def testAudioStream(filename):
    """Test if an audio stream definitely exists"""
    audioTestCommand = [
        Core.FFMPEG_BIN,
        "-i",
        filename,
        "-vn",
        "-f",
        "null",
        "-",
    ]
    try:
        checkOutput(audioTestCommand, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        return False
    else:
        return True


def getAudioDuration(filename):
    """Try to get duration of audio file as float, or False if not possible"""
    command = [Core.FFMPEG_BIN, "-i", filename]

    try:
        fileInfo = checkOutput(command, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as ex:
        fileInfo = ex.output
    except (FileNotFoundError, PermissionError):
        # ffmpeg is possibly not installed
        return False

    try:
        info = fileInfo.decode("utf-8").split("\n")
    except UnicodeDecodeError as e:
        log.error("Unicode error:", str(e))
        return False

    for line in info:
        if "Duration" in line:
            d = line.split(",")[0]
            d = d.split(" ")[3]
            d = d.split(":")
            duration = float(d[0]) * 3600 + float(d[1]) * 60 + float(d[2])
            break
    else:
        # String not found in output
        return False
    return duration


def readAudioFile(filename, videoWorker):
    """
    Creates the completeAudioArray given to components
    and used to draw the classic visualizer.
    """
    duration = getAudioDuration(filename)
    if not duration:
        log.error(f"Audio file {filename} doesn't exist or unreadable.")
        return

    command = [
        Core.FFMPEG_BIN,
        "-i",
        filename,
        "-f",
        "s16le",
        "-acodec",
        "pcm_s16le",
        "-ar",
        "44100",  # ouput will have 44100 Hz
        "-ac",
        "1",  # mono (set to '2' for stereo)
        "-",
    ]
    in_pipe = openPipe(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        bufsize=10**8,
    )

    completeAudioArray = numpy.empty(0, dtype="int16")

    progress = 0
    lastPercent = None
    while True:
        if Core.canceled:
            return
        # read 2 seconds of audio
        progress += 4
        raw_audio = in_pipe.stdout.read(88200 * 4)
        if len(raw_audio) == 0:
            break
        audio_array = numpy.frombuffer(raw_audio, dtype="int16")
        completeAudioArray = numpy.append(completeAudioArray, audio_array)

        percent = int(100 * (progress / duration))
        if percent >= 100:
            percent = 100

        if lastPercent != percent:
            string = "Loading audio file: " + str(percent) + "%"
            videoWorker.progressBarSetText.emit(string)
            videoWorker.progressBarUpdate.emit(percent)

        lastPercent = percent

    in_pipe.kill()
    in_pipe.wait()

    # add 0s the end
    completeAudioArrayCopy = numpy.zeros(len(completeAudioArray) + 44100, dtype="int16")
    completeAudioArrayCopy[: len(completeAudioArray)] = completeAudioArray
    completeAudioArray = completeAudioArrayCopy

    return (completeAudioArray, duration)


def exampleSound(style="white", extra="apulsator=offset_l=0.35:offset_r=0.67"):
    """Help generate an example sound for use in creating a preview"""

    if style == "white":
        src = "-2+random(0)"
    elif style == "freq":
        src = "sin(1000*t*PI*t)"
    elif style == "wave":
        src = "sin(random(0)*2*PI*t)*tan(random(0)*2*PI*t)"
    elif style == "stereo":
        src = "0.1*sin(2*PI*(360-2.5/2)*t) | 0.1*sin(2*PI*(360+2.5/2)*t)"

    return "aevalsrc='%s', %s%s" % (src, extra, ", " if extra else "")


def checkFfmpegVersion():
    try:
        with open(os.devnull, "w") as f:
            ffmpegVers = checkOutput([Core.FFMPEG_BIN, "-version"], stderr=f)
        ffmpegVers = str(ffmpegVers).split()[2].split(".", 1)[0]
        if ffmpegVers.startswith("n"):
            ffmpegVers = ffmpegVers[1:]
        versionNum = int(ffmpegVers)
        goodVersion = versionNum > 3
    except Exception:
        versionNum = -1
        goodVersion = False
    return goodVersion, versionNum



from pytest import fixture
from pytestqt import qtbot
from avp.command import Command
from avp.toolkit import blockSignals, rgbFromString
from . import command


@fixture
def gotWarning():
    """Check if a function called log.warning"""
    import avp.toolkit.common as tk

    warning = False

    def gotWarning():
        nonlocal warning
        return warning

    class log:
        def warning(self, *args):
            nonlocal warning
            warning = True

    oldLog = tk.log
    tk.log = log()
    try:
        yield gotWarning
    finally:
        tk.log = oldLog


def test_blockSignals(qtbot, command):
    command.core.insertComponent(0, 0, command)
    comp = command.core.selectedComponents[0]
    assert comp.page.spinBox_scale.signalsBlocked() == False
    with blockSignals(comp.page.spinBox_scale):
        assert comp.page.spinBox_scale.signalsBlocked() == True
    assert comp.page.spinBox_scale.signalsBlocked() == False


def test_rgbFromString(gotWarning):
    assert rgbFromString("255,255,255") == (255, 255, 255)
    assert not gotWarning()


def test_rgbFromString_error(gotWarning):
    assert rgbFromString("255,255,256") == (255, 255, 255)
    assert gotWarning()



import pytest
from avp.toolkit.ffmpeg import createFfmpegCommand
from . import audioData, getTestDataPath, command


def test_readAudioFile_data(audioData):
    assert len(audioData[0]) == 218453


def test_readAudioFile_duration(audioData):
    assert audioData[1] == 3.95


@pytest.mark.parametrize("width, height", ((1920, 1080), (1280, 720)))
def test_createFfmpegCommand(command, width, height):
    command.settings.setValue("outputWidth", width)
    command.settings.setValue("outputHeight", height)
    ffmpegCmd = createFfmpegCommand("test.ogg", "/tmp", command.core.selectedComponents)
    assert ffmpegCmd == [
        "ffmpeg",
        "-loglevel",
        "info",
        "-thread_queue_size",
        "512",
        "-y",
        "-f",
        "rawvideo",
        "-vcodec",
        "rawvideo",
        "-s",
        "%sx%s" % (width, height),
        "-pix_fmt",
        "rgba",
        "-r",
        "30",
        "-t",
        "0.100",
        "-an",
        "-i",
        "-",
        "-t",
        "0.100",
        "-i",
        "test.ogg",
        "-map",
        "0:v",
        "-map",
        "1:a",
        "-vcodec",
        "libx264",
        "-acodec",
        "aac",
        "-b:v",
        "2500k",
        "-b:a",
        "192k",
        "-pix_fmt",
        "yuv420p",
        "-preset",
        "medium",
        "-f",
        "mp4",
        "/tmp",
    ]



"""
Worker thread created to export a video. It has a slot to begin export using
an input file, output path, and component list.

Signals are emitted to update MainWindow's progress bar, detail text, and preview.
A Command object takes the place of MainWindow while in commandline mode.

Export can be cancelled with cancel()
"""

from PyQt6 import QtCore, QtGui
from PyQt6.QtCore import pyqtSignal, pyqtSlot
from PIL import Image
from PIL.ImageQt import ImageQt

import numpy
import subprocess as sp
import sys
import os
import signal
import logging

from .libcomponent import ComponentError
from .toolkit import formatTraceback
from .toolkit.ffmpeg import (
    openPipe,
    readAudioFile,
    getAudioDuration,
    createFfmpegCommand,
)


log = logging.getLogger("AVP.VideoThread")


class Worker(QtCore.QObject):

    imageCreated = pyqtSignal("QImage")
    videoCreated = pyqtSignal()
    progressBarUpdate = pyqtSignal(int)
    progressBarSetText = pyqtSignal(str)
    encoding = pyqtSignal(bool)

    def __init__(self, parent, inputFile, outputFile, components):
        super().__init__()
        self.core = parent.core
        self.settings = parent.settings
        self.modules = parent.core.modules
        parent.createVideo.connect(self.createVideo)
        self.previewEnabled = type(parent.core).previewEnabled

        self.components = components
        self.outputFile = outputFile
        self.inputFile = inputFile

        self.hertz = 44100
        self.sampleSize = 1470  # 44100 / 30 = 1470
        self.canceled = False
        self.error = False

    def createFfmpegCommand(self, duration):
        try:
            ffmpegCommand = createFfmpegCommand(
                self.inputFile,
                self.outputFile,
                self.components,
                duration,
                "info" if log.getEffectiveLevel() < logging.WARNING else "error",
            )
        except sp.CalledProcessError as e:
            # FIXME video_thread should own this error signal, not components
            self.components[0]._error.emit(
                "Ffmpeg could not be found. Is it installed?", str(e)
            )
            self.error = True
            return

        if not ffmpegCommand:
            # FIXME video_thread should own this error signal, not components
            self.components[0]._error.emit(
                "The FFmpeg command could not be generated.", ""
            )
            log.critical(
                "Cancelling render process due to failure while generating the ffmpeg command."
            )
            self.failExport()
            return
        return ffmpegCommand

    def determineAudioLength(self):
        """
        Returns audio length which determines length of final video, or False if failure occurs
        """
        if any(
            [True if "pcm" in comp.properties() else False for comp in self.components]
        ):
            self.progressBarSetText.emit("Loading audio file...")
            audioFileTraits = readAudioFile(self.inputFile, self)
            if audioFileTraits is None:
                self.cancelExport()
                return False
            self.completeAudioArray, duration = audioFileTraits
            self.audioArrayLen = len(self.completeAudioArray)
        else:
            duration = getAudioDuration(self.inputFile)
            self.completeAudioArray = []
            self.audioArrayLen = int(
                ((duration * self.hertz) + self.hertz) - self.sampleSize
            )
        return duration

    def preFrameRender(self):
        """
        Initializes components that need to pre-compute stuff.
        Also prerenders "static" components like text and merges them if possible
        """
        self.staticComponents = {}
        self.compositeComponents = set()

        # Call preFrameRender on each component
        canceledByComponent = False
        initText = ", ".join(
            [
                "%s) %s" % (num, str(component))
                for num, component in enumerate(reversed(self.components))
            ]
        )
        print("Loaded Components:", initText)
        log.info("Calling preFrameRender for %s", initText)
        for compNo, comp in enumerate(reversed(self.components)):
            try:
                comp.preFrameRender(
                    audioFile=self.inputFile,
                    completeAudioArray=self.completeAudioArray,
                    audioArrayLen=self.audioArrayLen,
                    sampleSize=self.sampleSize,
                    progressBarUpdate=self.progressBarUpdate,
                    progressBarSetText=self.progressBarSetText,
                )
            except ComponentError:
                log.warning(
                    "#%s %s encountered an error in its preFrameRender method",
                    compNo,
                    comp,
                )

            compProps = comp.properties()
            if "error" in compProps or comp._lockedError is not None:
                self.cancel()
                self.canceled = True
                canceledByComponent = True
                compError = (
                    comp.error() if type(comp.error()) is tuple else (comp.error(), "")
                )
                errMsg = (
                    "Component #%s (%s) encountered an error!"
                    % (str(compNo), comp.name)
                    if comp.error() is None
                    else "Export cancelled by component #%s (%s): %s"
                    % (str(compNo), comp.name, compError[0])
                )
                log.error(errMsg)
                comp._error.emit(errMsg, compError[1])
                break
            if "static" in compProps:
                log.info("Saving static frame from #%s %s", compNo, comp)
                self.staticComponents[compNo] = comp.frameRender(0).copy()
            elif compNo > 0 and "composite" in compProps:
                self.compositeComponents.add(compNo)

        # Check if any errors occured
        log.debug("Checking if a component wishes to cancel the export...")
        if self.canceled:
            if canceledByComponent:
                log.error(
                    "Export cancelled by component #%s (%s): %s",
                    compNo,
                    comp.name,
                    (
                        "No message."
                        if comp.error() is None
                        else (
                            comp.error()
                            if type(comp.error()) is str
                            else comp.error()[0]
                        )
                    ),
                )
            self.cancelExport()

        # Merge static frames that can be merged to reduce workload
        def mergeConsecutiveStaticComponentFrames(self):
            log.info("Merging consecutive static component frames")
            for compNo in range(len(self.components)):
                if (
                    compNo not in self.staticComponents
                    or compNo + 1 not in self.staticComponents
                ):
                    continue
                self.staticComponents[compNo + 1] = Image.alpha_composite(
                    self.staticComponents.pop(compNo),
                    self.staticComponents[compNo + 1],
                )
                self.staticComponents[compNo] = None

        mergeConsecutiveStaticComponentFrames(self)

    def frameRender(self, audioI):
        """
        Renders a frame composited together from the frames returned by each component
        audioI is a multiple of self.sampleSize, which can be divided to determine frameNo
        """

        def err():
            self.closePipe()
            self.cancelExport()
            self.error = True
            msg = f"{comp.name} renderFrame({int(audioI / self.sampleSize)}) raised an exception."
            tb = formatTraceback()
            details = f"{e.__class__.__name__}: {str(e)}\n\n{tb}"
            log.critical(f"{msg}\n{details}")
            comp._error.emit(msg, details)

        bgI = int(audioI / self.sampleSize)
        frame = None
        for layerNo, comp in enumerate(reversed((self.components))):
            if self.canceled:
                break
            try:
                if layerNo in self.staticComponents:
                    if self.staticComponents[layerNo] is None:
                        # this layer was merged into a following layer
                        continue
                    # static component
                    if frame is None:  # bottom-most layer
                        frame = self.staticComponents[layerNo]
                    else:
                        frame = Image.alpha_composite(
                            frame, self.staticComponents[layerNo]
                        )

                elif layerNo in self.compositeComponents:
                    # component that uses previous frame to draw
                    frame = Image.alpha_composite(frame, comp.frameRender(bgI, frame))
                else:
                    # animated component
                    if frame is None:  # bottom-most layer
                        frame = comp.frameRender(bgI)
                    else:
                        frame = Image.alpha_composite(frame, comp.frameRender(bgI))
            except Exception as e:
                err()
        return frame

    def showPreview(self, frame):
        """
        Receives a final frame that will be piped to FFmpeg,
        adds it to the MainWindow for the live preview
        """
        # We must store a reference to this QImage
        # or else Qt will garbage-collect it on the C++ side
        self.latestPreview = ImageQt(frame)
        self.imageCreated.emit(QtGui.QImage(self.latestPreview))

    @pyqtSlot()
    def createVideo(self):
        """
        1. Determine length of final video
        2. Call preFrameRender on each component
        3. Create the main FFmpeg command
        4. Open the out_pipe to FFmpeg process
        5. Iterate over the audio data array and call frameRender on the components to get frames
        6. Close the out_pipe
        7. Call postFrameRender on each component
        """
        log.debug("Video worker received signal to createVideo")
        log.debug("Video thread id: {}".format(int(QtCore.QThread.currentThreadId())))
        self.encoding.emit(True)
        self.extraAudio = []
        self.width = int(self.settings.value("outputWidth"))
        self.height = int(self.settings.value("outputHeight"))

        # Set core.Core.canceled to False and call .reset() on each component
        self.reset()
        # Initialize progress bar to 0
        progressBarValue = 0
        self.progressBarUpdate.emit(progressBarValue)

        # Determine longest length of audio which will be the final video's duration
        log.debug("Determining length of audio...")
        duration = self.determineAudioLength()
        if not duration:
            return

        # Call preFrameRender on each component to perform initialization
        self.progressBarUpdate.emit(0)
        self.progressBarSetText.emit("Starting components...")
        self.preFrameRender()
        if self.canceled:
            return

        # Create FFmpeg command
        ffmpegCommand = self.createFfmpegCommand(duration)
        if not ffmpegCommand:
            return
        cmd = " ".join(ffmpegCommand)
        print("###### FFMPEG COMMAND ######\n%s" % cmd)
        print("############################")
        log.info(cmd)

        # Open pipe to FFmpeg
        log.info("Opening pipe to FFmpeg")
        try:
            self.out_pipe = openPipe(
                ffmpegCommand,
                stdin=sp.PIPE,
                stdout=sys.stdout,
                stderr=sys.stdout,
            )
        except sp.CalledProcessError:
            log.critical("Out_Pipe to FFmpeg couldn't be created!", exc_info=True)
            raise

        # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
        # START CREATING THE VIDEO
        # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
        progressBarValue = 0
        self.progressBarUpdate.emit(progressBarValue)
        # Begin piping into ffmpeg!
        self.progressBarSetText.emit("Exporting video...")
        for audioI in range(0, self.audioArrayLen, self.sampleSize):
            if self.canceled:
                break
            # fetch the next frame & add to the FFmpeg pipe
            frame = self.frameRender(audioI)

            # Update live preview
            if self.previewEnabled:
                self.showPreview(frame)

            try:
                self.out_pipe.stdin.write(frame.tobytes())
            except Exception:
                break

            # increase progress bar value
            completion = (audioI / self.audioArrayLen) * 100
            if progressBarValue + 1 <= completion:
                progressBarValue = numpy.floor(completion).astype(int)
                msg = "Exporting video: %s%%" % str(int(progressBarValue))
                self.progressBarUpdate.emit(progressBarValue)
                self.progressBarSetText.emit(msg)

        # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
        # Finished creating the video!
        # =~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

        self.closePipe()

        for comp in reversed(self.components):
            comp.postFrameRender()

        if self.canceled:
            print("Export Canceled")
            try:
                os.remove(self.outputFile)
            except Exception:
                pass
            self.progressBarUpdate.emit(0)
            self.progressBarSetText.emit("Export Canceled")
        else:
            if self.error:
                self.failExport()
            else:
                print("\nExport Complete")
                self.progressBarUpdate.emit(100)
                self.progressBarSetText.emit("Export Complete")

        self.error = False
        self.canceled = False
        self.encoding.emit(False)
        self.videoCreated.emit()

    def closePipe(self):
        try:
            self.out_pipe.stdin.close()
        except (BrokenPipeError, OSError):
            log.debug("Broken pipe to FFmpeg!")
        if self.out_pipe.stderr is not None:
            log.error(self.out_pipe.stderr.read())
            self.out_pipe.stderr.close()
            self.error = True
        self.out_pipe.wait()

    def cancelExport(self, message="Export Canceled"):
        self.progressBarUpdate.emit(0)
        self.progressBarSetText.emit(message)
        self.encoding.emit(False)
        self.videoCreated.emit()

    def failExport(self):
        self.cancelExport("Export Failed")

    def updateProgress(self, pStr, pVal):
        self.progressBarValue.emit(pVal)
        self.progressBarSetText.emit(pStr)

    def cancel(self):
        self.canceled = True
        self.core.cancel()

        for comp in self.components:
            comp.cancel()

        try:
            self.out_pipe.send_signal(signal.SIGTERM)
        except Exception:
            pass

    def reset(self):
        self.core.reset()
        self.canceled = False
        for comp in self.components:
            comp.reset()
